# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import scipy.constants as constants
import numpy as np

from ..timing import function_timer, GlobalTimers

from ..op import Operator

from .. import qarray as qa

from .._libtoast import (
    cov_accum_diag,
    cov_accum_zmap,
    cov_accum_diag_hits,
    cov_accum_diag_invnpp,
    scan_map_float64,
    scan_map_float32,
    apply_flags_to_pixels,
)

from ..map import DistPixels


class OpAccumDiag(Operator):
    """Operator which accumulates the diagonal covariance and noise weighted map.

    This operator requires that the local pointing matrix has already been
    computed.  Each process has local pieces of the map products.  This
    operator can optionally accumulate any combination of the hit map,
    the white noise weighted map, and the diagonal inverse pixel covariance.

    NOTE:  The input DistPixels objects (noise weighted map, hit map, and
    inverse covariance) are purposefully NOT set to zero at the start.  This
    allows accumulating multiple instances of the distributed data.  This might
    be needed (for example) if all desired timestream data does not fit into
    memory.  You should manually clear the pixel domain objects before
    accumulation if desired.

    Args:
        zmap (DistPixels):  (optional) the noise weighted map to accumulate.
        hits (DistPixels):  (optional) the hits to accumulate.
        invnpp (DistPixels):  (optional) the diagonal covariance matrix.
        detweights (dictionary): individual noise weights to use for each
            detector.
        name (str): the name of the cache object (<name>_<detector>) to
            use for the detector timestream.  If None, use the TOD.
        flag_name (str): the name of the cache object (<flag_name>_<detector>) to
            use for the detector flags.  If None, use the TOD.
        flag_mask (int): the integer bit mask (0-255) that should be
            used with the detector flags in a bitwise AND.
        common_flag_name (str): the name of the cache object
            (<common_flag_name>_<detector>) to use for the common flags.
            If None, use the TOD.
        common_flag_mask (int): the integer bit mask (0-255) that should be
            used with the common flags in a bitwise AND.
        apply_flags (bool): whether to apply flags to the pixel numbers.
        pixels (str): the name of the cache object (<pixels>_<detector>)
            containing the pixel indices to use.
        weights (str): the name of the cache object (<weights>_<detector>)
            containing the pointing weights to use.
        detectors (iterable):  List of detectors to process.  If None, use
            all detectors.
    """

    def __init__(
        self,
        zmap=None,
        hits=None,
        invnpp=None,
        detweights=None,
        name=None,
        flag_name=None,
        flag_mask=255,
        common_flag_name=None,
        common_flag_mask=255,
        pixels="pixels",
        weights="weights",
        apply_flags=True,
        detectors=None,
    ):

        self._flag_name = flag_name
        self._flag_mask = flag_mask
        self._common_flag_name = common_flag_name
        self._common_flag_mask = common_flag_mask
        self._apply_flags = apply_flags

        self._name = name
        self._pixels = pixels
        self._weights = weights
        self._detweights = detweights
        self._detectors = detectors

        # Ensure that the 3 different DistPixel objects have the same number
        # of pixels.

        self._nsub = None
        self._subsize = None
        self._nnz = None

        self._do_z = False
        self._do_hits = False
        self._do_invn = False

        self._zmap = zmap
        self._hits = hits
        self._invnpp = invnpp

        self._globloc = None

        if zmap is not None:
            self._do_z = True
            self._nsub = zmap.nsubmap
            self._subsize = zmap.npix_submap
            self._nnz = zmap.nnz
            if self._globloc is None:
                self._globloc = self._zmap

        if hits is not None:
            self._do_hits = True
            if hits.nnz != 1:
                raise RuntimeError("Hit map should always have NNZ == 1")
            if self._nsub is None:
                self._nsub = hits.nsubmap
                self._subsize = hits.npix_submap
            else:
                if self._nsub != hits.nsubmap:
                    raise RuntimeError(
                        "All pixel domain objects must have the same number "
                        "of local submaps."
                    )
                if self._subsize != hits.npix_submap:
                    raise RuntimeError(
                        "All pixel domain objects must have the same submap size."
                    )
            if self._globloc is None:
                self._globloc = self._hits

        if invnpp is not None:
            self._do_invn = True
            block = invnpp.nnz
            blocknnz = int(((np.sqrt(8 * block) - 1) / 2) + 0.5)
            if self._nsub is None:
                self._nsub = invnpp.nsubmap
                self._subsize = invnpp.npix_submap
                self._nnz = blocknnz
            else:
                if self._nsub != invnpp.nsubmap:
                    raise RuntimeError(
                        "All pixel domain objects must have the same submap size."
                    )
                if self._subsize != invnpp.npix_submap:
                    raise RuntimeError(
                        "All pixel domain objects must have the same submap size."
                    )
                if self._nnz is None:
                    self._nnz = blocknnz
                elif self._nnz != blocknnz:
                    raise RuntimeError(
                        "All pixel domain objects must have the same submap size."
                    )
            if self._globloc is None:
                self._globloc = self._invnpp

        if self._nnz is None:
            # this means we only have a hit map
            self._nnz = 1

        if self._do_invn and (not self._do_hits):
            raise RuntimeError(
                "When accumulating the diagonal pixel covariance, you must "
                "also accumulate the hit map"
            )

        if self._do_z and (self._do_hits != self._do_invn):
            raise RuntimeError(
                "When accumulating the noise weighted map, you must accumulate "
                "either both the hits and covariance or neither."
            )

        # We call the parent class constructor, which currently does nothing
        super().__init__()

    @function_timer
    def exec(self, data):
        """Iterate over all observations and detectors and accumulate.

        Args:
            data (toast.Data): The distributed data.
        """
        # the two-level pytoast communicator
        comm = data.comm
        # the global communicator
        cworld = comm.comm_world
        # the communicator within the group
        cgroup = comm.comm_group
        # the communicator with all processes with
        # the same rank within their group
        crank = comm.comm_rank
        gt = GlobalTimers.get()

        for obs in data.obs:
            tod = obs["tod"]

            nsamp = tod.local_samples[1]

            commonflags = None
            if self._apply_flags:
                commonflags = tod.local_common_flags(self._common_flag_name).copy()

            for det in tod.local_dets:
                if self._detectors is not None and det not in self._detectors:
                    continue

                # get the pixels and weights from the cache

                pixelsname = "{}_{}".format(self._pixels, det)
                weightsname = "{}_{}".format(self._weights, det)
                pixels = tod.cache.reference(pixelsname)
                weights = tod.cache.reference(weightsname)

                cachename = None
                signal = None

                if self._do_z:
                    signal = tod.local_signal(det, self._name)

                # get flags

                if self._apply_flags:
                    gt.start("OpAccumDiag.exec.apply_flags")
                    # Don't change the cached pixel numbers
                    pixels = pixels.copy()
                    detflags = tod.local_flags(det, self._flag_name)
                    apply_flags_to_pixels(
                        commonflags,
                        self._common_flag_mask,
                        detflags,
                        self._flag_mask,
                        pixels,
                    )
                    gt.stop("OpAccumDiag.exec.apply_flags")

                # local pointing

                gt.start("OpAccumDiag.exec.global_to_local")
                sm, lpix = self._globloc.global_to_local(pixels)
                gt.stop("OpAccumDiag.exec.global_to_local")

                detweight = 1.0

                if self._detweights is not None:
                    if det not in self._detweights.keys():
                        raise RuntimeError(
                            "no detector weights found for {}".format(det)
                        )
                    detweight = self._detweights[det]
                    if detweight == 0:
                        continue

                # Now call the correct accumulation operator depending
                # on which input pixel objects were given.

                if self._do_invn and self._do_z:
                    invnpp = self._invnpp.flatdata
                    if invnpp is None:
                        invnpp = np.empty(shape=0, dtype=np.float64)
                    zmap = self._zmap.flatdata
                    if zmap is None:
                        zmap = np.empty(shape=0, dtype=np.float64)
                    hits = self._hits.flatdata
                    if hits is None:
                        hits = np.empty(shape=0, dtype=np.int64)
                    cov_accum_diag(
                        self._nsub,
                        self._subsize,
                        self._nnz,
                        sm,
                        lpix,
                        weights.reshape(-1),
                        detweight,
                        signal,
                        invnpp,
                        hits,
                        zmap,
                    )

                elif self._do_invn:
                    invnpp = self._invnpp.flatdata
                    if invnpp is None:
                        invnpp = np.empty(shape=0, dtype=np.float64)
                    hits = self._hits.flatdata
                    if hits is None:
                        hits = np.empty(shape=0, dtype=np.int64)
                    cov_accum_diag_invnpp(
                        self._nsub,
                        self._subsize,
                        self._nnz,
                        sm,
                        lpix,
                        weights.reshape(-1),
                        detweight,
                        invnpp,
                        hits,
                    )

                elif self._do_z:
                    zmap = self._zmap.flatdata
                    if zmap is None:
                        zmap = np.empty(shape=0, dtype=np.float64)
                    cov_accum_zmap(
                        self._nsub,
                        self._subsize,
                        self._nnz,
                        sm.astype(np.int64),
                        lpix.astype(np.int64),
                        weights.reshape(-1),
                        detweight,
                        signal,
                        zmap,
                    )

                elif self._do_hits:
                    hits = self._hits.flatdata
                    if hits is None:
                        hits = np.empty(shape=0, dtype=np.int64)
                    cov_accum_diag_hits(
                        self._nsub,
                        self._subsize,
                        self._nnz,
                        sm.astype(np.int64),
                        lpix.astype(np.int64),
                        hits,
                    )

                # print("det {}:".format(det))
                # if self._zmap is not None:
                #     print(self._zmap.data)
                # if self._hits is not None:
                #     print(self._hits.data)
                # if self._invnpp is not None:
                #     print(self._invnpp.data)

        return


class OpScanScale(Operator):
    """Operator which scales sky signal by weights from a map.

    Local pixels should already exist.

    Args:
        distmap (DistPixels): the distributed map domain data.
        pixels (str): the name of the cache object (<pixels>_<detector>)
            containing the pixel indices to use.
        name (str): scale data in cache with name <name>_<detector>.

    """

    def __init__(self, distmap=None, pixels="pixels", name=None):
        # Call the parent class constructor
        super().__init__()
        self.map = distmap
        self.pixels = pixels
        self.name = name

    def exec(self, data):
        """Scale signal by scanning weights from the map.

        This loops over all observations and detectors and uses the pointing
        matrix to project the distributed map into a timestream.

        Args:
            data (toast.Data): The distributed data.

        Returns:
            None

        """
        for obs in data.obs:
            tod = obs["tod"]
            for det in tod.local_dets:
                # get the pixels and weights from the cache
                pixelsname = "{}_{}".format(self.pixels, det)
                pixels = tod.cache.reference(pixelsname)
                sm, lpix = self.map.global_to_local(pixels)
                ref = tod.local_signal(det, self.name)
                weighted = np.zeros(pixels.size)

                maptype = np.dtype(self.map.dtype)
                if maptype.char == "d":
                    scan_map = scan_map_float64
                elif maptype.char == "f":
                    scan_map = scan_map_float32
                else:
                    raise RuntimeError(
                        "Scanning weights from a map only supports float32 and float64 maps"
                    )
                # We pass the signal to be scaled in place of the pointing weights
                # The returned TOD is already TOD x weigths
                scan_map(
                    self.map.npix_submap,
                    1,
                    sm,
                    lpix,
                    self.map.flatdata,
                    ref.astype(np.float64),
                    weighted,
                )
                ref[:] = weighted
                del ref
                del pixels
        return


class OpScanMask(Operator):
    """Operator which scans a mask and sets the TOD flags accordingly

    Local pixels should already exist.

    Args:
        distmap (DistPixels): the distributed mask.
        pixels (str): the name of the cache object (<pixels>_<detector>)
            containing the pixel indices to use.
        name (str): scale data in cache with name <name>_<detector>.
    """

    def __init__(self, distmap=None, pixels="pixels", flags=None, flagmask=1):
        # Call the parent class constructor
        super().__init__()
        self.map = distmap
        self.pixels = pixels
        self.flags = flags
        self.flagmask = flagmask

    def exec(self, data):
        """Scan mask values from a map and update the detector flags.
        Positive (non-zero) mask values indicate usable sky pixels.

        Args:
            data (toast.Data): The distributed data.

        Returns:
            None
        """
        for obs in data.obs:
            tod = obs["tod"]
            for det in tod.local_dets:
                # get the pixels and weights from the cache
                pixelsname = "{}_{}".format(self.pixels, det)
                pixels = tod.cache.reference(pixelsname)
                nsamp = pixels.size
                sm, lpix = self.map.global_to_local(pixels)

                maptype = np.dtype(self.map.dtype)
                if maptype.char == "d":
                    scan_map = scan_map_float64
                elif maptype.char == "f":
                    scan_map = scan_map_float32
                else:
                    raise RuntimeError(
                        "Scanning weights from a map only supports float32 and float64 maps"
                    )
                # We pass the signal to be scaled in place of the pointing weights
                # The returned TOD is already TOD x weigths
                weights = np.ones(nsamp, dtype=np.float64)
                masktod = np.zeros(nsamp, dtype=np.float64)
                scan_map(
                    self.map.npix_submap,
                    1,
                    sm,
                    lpix,
                    self.map.flatdata,
                    weights,
                    masktod,
                )
                flags = tod.local_flags(det, self.flags)
                flags[masktod < 0.5] |= self.flagmask
        return


def array_dot(u, v):
    """Dot product of each row of two 2D arrays"""
    return np.sum(u * v, axis=1).reshape((-1, 1))


@function_timer
def dipole(pntg, vel=None, solar=None, cmb=2.72548, freq=0):
    """Compute a dipole timestream.

    This uses detector pointing, telescope velocity and the solar system
    motion to compute the observed dipole.  It is assumed that the detector
    pointing, the telescope velocity vectors, and the solar system velocity
    vector are all in the same coordinate system.

    Args:
        pntg (array): the 2D array of quaternions of detector pointing.
        vel (array): 2D array of velocity vectors relative to the solar
            system barycenter.  if None, return only the solar system dipole.
            Units are km/s
        solar (array): a 3 element vector containing the solar system velocity
            vector relative to the CMB rest frame.  Units are km/s.
        cmb (float): CMB monopole in Kelvin.  Default value from Fixsen
            2009 (see arXiv:0911.1955)
        freq (float): optional observing frequency in Hz (NOT GHz).

    Returns:
        (array):  detector dipole timestream.

    """
    zaxis = np.array([0, 0, 1], dtype=np.float64)
    nsamp = pntg.shape[0]

    inv_light = 1.0e3 / constants.speed_of_light

    if (vel is not None) and (solar is not None):
        # relativistic addition of velocities

        solar_speed = np.sqrt(np.sum(solar * solar, axis=0))

        vpar = (array_dot(vel, solar) / solar_speed ** 2) * solar
        vperp = vel - vpar

        vdot = 1.0 / (1.0 + array_dot(solar, vel) * inv_light ** 2)
        invgamma = np.sqrt(1.0 - (solar_speed * inv_light) ** 2)

        vpar += solar
        vperp *= invgamma

        v = vdot * (vpar + vperp)
    elif solar is not None:
        v = np.tile(solar, nsamp).reshape((-1, 3))
    elif vel is not None:
        v = vel.copy()

    speed = np.sqrt(array_dot(v, v))
    v /= speed

    beta = inv_light * speed.flatten()

    direct = qa.rotate(pntg, zaxis)

    dipoletod = None
    if freq == 0:
        inv_gamma = np.sqrt(1.0 - beta ** 2)
        num = 1.0 - beta * np.sum(v * direct, axis=1)
        dipoletod = cmb * (inv_gamma / num - 1.0)
    else:
        # Use frequency for quadrupole correction
        fx = constants.h * freq / (constants.k * cmb)
        fcor = (fx / 2) * (np.exp(fx) + 1) / (np.exp(fx) - 1)
        bt = beta * np.sum(v * direct, axis=1)
        dipoletod = cmb * (bt + fcor * bt ** 2)

    return dipoletod



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import numpy as np

from .mpi import MPITestCase

from ..tod import TODCache

from ._helpers import create_outdir


class TODTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        rank = 0
        nproc = 1
        if self.comm is not None:
            rank = self.comm.rank
            nproc = self.comm.size
        self.dets = ["1a", "1b", "2a", "2b"]
        self.mynsamp = 10
        self.myoff = self.mynsamp * rank
        self.totsamp = self.mynsamp * nproc
        self.tod = TODCache(self.comm, self.dets, self.totsamp)
        self.rms = 10.0
        self.pntgvec = np.ravel(np.random.random((self.mynsamp, 4))).reshape(-1, 4)
        self.pflagvec = np.random.uniform(low=0, high=1, size=self.mynsamp).astype(
            np.uint8, copy=True
        )

        self.datavec = np.random.normal(loc=0.0, scale=self.rms, size=self.mynsamp)
        self.flagvec = np.random.uniform(low=0, high=1, size=self.mynsamp).astype(
            np.uint8, copy=True
        )
        self.tod.write_times(stamps=np.arange(self.mynsamp))
        for d in self.dets:
            self.tod.write_common_flags(local_start=0, flags=self.pflagvec)
            self.tod.write_flags(detector=d, local_start=0, flags=self.flagvec)
            self.tod.write(detector=d, local_start=0, data=self.datavec)
            self.tod.write_pntg(detector=d, local_start=0, data=self.pntgvec)

    def tearDown(self):
        pass

    def test_props(self):
        comm = self.tod.mpicomm
        if comm is None:
            print(self.tod, flush=True)
        else:
            for p in range(comm.size):
                if p == comm.rank:
                    print(self.tod, flush=True)
                comm.barrier()

        self.assertEqual(sorted(self.tod.detectors), sorted(self.dets))
        self.assertEqual(sorted(self.tod.local_dets), sorted(self.dets))
        self.assertEqual(self.tod.total_samples, self.totsamp)
        self.assertEqual(self.tod.local_samples[0], self.myoff)
        self.assertEqual(self.tod.local_samples[1], self.mynsamp)
        return

    def test_read(self):
        common = self.tod.read_common_flags(local_start=0, n=self.mynsamp)
        for d in self.dets:
            data = self.tod.read(detector=d, local_start=0, n=self.mynsamp)
            flags = self.tod.read_flags(detector=d, local_start=0, n=self.mynsamp)
            np.testing.assert_equal(flags, self.flagvec)
            np.testing.assert_equal(common, self.pflagvec)
            np.testing.assert_almost_equal(data, self.datavec)
        return

    def test_cached_read(self):
        common = self.tod.local_common_flags()
        for d in self.dets:
            data = self.tod.local_signal(d)
            flags = self.tod.local_flags(d)
            np.testing.assert_equal(flags, self.flagvec)
            np.testing.assert_equal(common, self.pflagvec)
            np.testing.assert_almost_equal(data, self.datavec)

        # Then we can use the cached TOD

        common = self.tod.cache.reference("common_flags")
        for d in self.dets:
            data = self.tod.cache.reference("signal_" + d)
            flags = self.tod.cache.reference("flags_" + d)
            np.testing.assert_equal(flags, self.flagvec)
            np.testing.assert_equal(common, self.pflagvec)
            np.testing.assert_almost_equal(data, self.datavec)

        return

    def test_read_pntg(self):
        for d in self.dets:
            pntg = self.tod.read_pntg(detector=d, local_start=0, n=self.mynsamp)
            np.testing.assert_almost_equal(pntg, self.pntgvec)
        return

    def test_local_intervals(self):
        local_intervals = self.tod.local_intervals(None)
        self.assertEqual(
            self.mynsamp, local_intervals[0].last - local_intervals[0].first + 1
        )
        return

    def test_local_signal(self):
        for d in self.dets:
            data = self.tod.local_signal(d)
            np.testing.assert_almost_equal(data, self.datavec)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from scipy.constants import c

from .. import qarray as qa

from ..timing import function_timer


cinv = 1e3 / c  # Inverse light speed in km / s ( the assumed unit for velocity )

xaxis, yaxis, zaxis = np.eye(3, dtype=np.float64)

# J2000 coordinate transforms

# RA, DEC to galactic coordinates

coordmat_J2000radec2gal = np.array(
    [
        -0.054875539726,
        -0.873437108010,
        -0.483834985808,
        0.494109453312,
        -0.444829589425,
        0.746982251810,
        -0.867666135858,
        -0.198076386122,
        0.455983795705,
    ]
).reshape([3, 3])
quat_equ2gal = qa.from_rotmat(coordmat_J2000radec2gal)

# RA, DEC to (geocentric) ecliptic coordinates

coordmat_J2000radec2ecl = np.array(
    [
        1.0,
        0.0,
        0.0,
        0.0,
        0.917482062069182,
        0.397777155931914,
        0.0,
        -0.397777155931914,
        0.917482062069182,
    ]
).reshape([3, 3])
quat_equ2ecl = qa.from_rotmat(coordmat_J2000radec2ecl)

# Ecliptic coordinates (geocentric) to galactic
# (use the same rotation as HEALPix, to avoid confusion)
coordmat_J2000ecl2gal = np.array(
    [
        -0.054882486,
        -0.993821033,
        -0.096476249,
        0.494116468,
        -0.110993846,
        0.862281440,
        -0.867661702,
        -0.000346354,
        0.497154957,
    ]
).reshape([3, 3])
quat_ecl2gal = qa.from_rotmat(coordmat_J2000ecl2gal)


@function_timer
def aberrate(quat, vel, inplace=True):
    """Apply velocity aberration to the orientation quaternions.

    Args:
        quat (float):  Normalized quaternions.
        vel (float):  Telescope velocity with respect to the signal rest frame.

    Returns:
        Corrected quaternions either in the orinal container or a 2D ndarray.

    """
    vec = qa.rotate(quat, zaxis)
    abvec = np.cross(vec, vel)
    lens = np.linalg.norm(abvec, axis=1)
    ang = lens * cinv
    abvec /= np.tile(lens, (3, 1)).T  # Normalize for direction
    abquat = qa.rotation(abvec, -ang)

    if inplace:
        quat[:] = qa.mult(abquat, quat)
        return

    return qa.mult(abquat, quat)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import numpy as np

from .mpi import MPITestCase

from ..tod.applygain import write_calibration_file, OpApplyGain

from ..tod.tod import TODCache

from ._helpers import create_outdir, create_distdata, boresight_focalplane


class TestApplyGain(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)
        self.data = create_distdata(self.comm, obs_per_group=1)
        self.gain = {
            "TIME": np.arange(10),
            "1a": 2 * np.ones(10, dtype=np.float32),
            "1b": 3 * np.ones(10, dtype=np.float32),
        }
        self.gain["1b"][5:] = 0

    def test_write_calibration_file(self):
        # Only one process should write, otherwise they will
        # stomp on each other.
        outfile = os.path.join(self.outdir, "test_cal.fits")
        if (self.comm is None) or (self.comm.rank == 0):
            write_calibration_file(outfile, self.gain)

    def test_op_applygain(self):
        # Two detectors, default properties
        self.ndet = 2
        self.dets = ["1a", "1b"]
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(self.ndet)

        # Pick some number of samples per observation
        self.samples_per_obs = 10

        # (there is only one observation per group)
        self.data.obs[0]["tod"] = TODCache(
            self.data.comm.comm_group, self.dets, self.samples_per_obs
        )

        # generate input data of ones for each of the 2 channels and 10 timestamps
        # from 1 to 10, no need for pointing

        for obs in self.data.obs:
            tod = obs["tod"]
            # create timestamps for our local samples
            time_stamps = np.arange(tod.local_samples[1], dtype=np.float64)
            time_stamps += tod.local_samples[0]
            tod.write_times(stamps=time_stamps)
            # write our local data
            for det in tod.local_dets:
                tod.write(detector=det, data=np.ones(tod.local_samples[1]))

        op_apply_gain = OpApplyGain(self.gain)
        op_apply_gain.exec(self.data)

        # compare calibrated timelines

        for obs in self.data.obs:
            tod = obs["tod"]
            locoff = tod.local_samples[0]
            nloc = tod.local_samples[1]
            # check our local pieces of data
            check = 2.0 * np.ones(nloc)
            np.testing.assert_allclose(check, tod.read(detector="1a"))
            check = np.zeros(self.samples_per_obs)
            check[0:5] = 3
            localcheck = check[locoff : locoff + nloc]
            np.testing.assert_allclose(localcheck, tod.read(detector="1b"))



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from ..mpi import MPI

from ..op import Operator

from ..utils import Logger

from ..timing import function_timer


class OpMemoryCounter(Operator):
    """Compute total memory used by TOD objects.

    Operator which loops over the TOD objects and computes the total
    amount of memory allocated.

    Args:
        silent (bool):  Only count and return the memory without
            printing.
        *other_caching_objects:  Additional objects that have a cache
            member and user wants to include in the total counts
            (e.q. DistPixels objects).

    """

    def __init__(self, *other_caching_objects, silent=False):
        self._silent = silent
        self._objects = []
        for obj in other_caching_objects:
            self._objects.append(obj)
        super().__init__()

    @function_timer
    def exec(self, data):
        """
        Count the memory

        Args:
            data (toast.Data): The distributed data.

        """
        log = Logger.get()
        # the two-level pytoast communicator
        comm = data.comm
        # the global communicator
        cworld = comm.comm_world
        # the communicator within the group
        cgroup = comm.comm_group

        tot_task = 0
        for obj in self._objects:
            try:
                tot_task += obj.cache.report(silent=True)
            except:
                pass
            try:
                tot_task += obj._cache.report(silent=True)
            except:
                pass

        for obs in data.obs:
            tod = obs["tod"]
            tot_task += tod.cache.report(silent=True)

        tot_group = tot_task
        tot_world = tot_task
        tot_task_max = tot_task
        tot_group_max = tot_task
        if cworld is not None:
            tot_group = cgroup.allreduce(tot_task, op=MPI.SUM)
            tot_world = cworld.allreduce(tot_task, op=MPI.SUM)
            tot_task_max = cworld.allreduce(tot_task, op=MPI.MAX)
            tot_group_max = cgroup.allreduce(tot_group, op=MPI.MAX)

        if (not self._silent) and (cworld is None or cworld.rank == 0):
            msg = "Memory usage statistics:\n\
                - Max memory (task): {:.2f} GB\n\
                - Max memory (group): {:.2f} GB\n\
                Total memory: {:.2f} GB\n\
                ".format(
                (tot_task_max / 2 ** 30),
                (tot_group_max / 2 ** 30),
                (tot_world / 2 ** 30),
            )
            log.info(msg)

        return tot_world



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os

import numpy as np

from ..tod import OpPolyFilter, AnalyticNoise, OpSimNoise, Interval
from ..todmap import TODHpixSpiral

from ._helpers import create_outdir, create_distdata, boresight_focalplane


class OpPolyFilterTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # Create one observation per group, and each observation will have
        # one detector per process and a single chunk.  Data within an
        # observation is distributed by detector.

        self.data = create_distdata(self.comm, obs_per_group=1)
        self.ndet = self.data.comm.group_size
        self.rate = 20.0

        # Create detectors with a range of knee frequencies.
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(
            self.ndet,
            samplerate=self.rate,
            net=10.0,
            fmin=1.0e-5,
            fknee=np.linspace(0.01, 0.19, num=self.ndet),
        )

        # Fitting order
        self.order = 5

        # Total Samples
        self.totsamp = 100000

        # Populate the observations (one per group)

        tod = TODHpixSpiral(
            self.data.comm.comm_group,
            dquat,
            self.totsamp,
            detranks=self.data.comm.group_size,
            firsttime=0.0,
            rate=self.rate,
            nside=512,
        )

        # Construct an analytic noise model for the detectors

        nse = AnalyticNoise(
            rate=drate,
            fmin=dfmin,
            detectors=dnames,
            fknee=dfknee,
            alpha=dalpha,
            NET=dnet,
        )

        intervals = []
        interval_len = 1000
        for istart in range(0, self.totsamp, interval_len):
            istop = min(istart + interval_len, self.totsamp)
            intervals.append(
                Interval(
                    start=istart / self.rate,
                    stop=istop / self.rate,
                    first=istart,
                    last=istop - 1,
                )
            )

        self.data.obs[0]["tod"] = tod
        self.data.obs[0]["noise"] = nse
        self.data.obs[0]["intervals"] = intervals

    def test_filter(self):
        # generate timestreams
        op = OpSimNoise()
        op.exec(self.data)

        # Replace the noise with a polynomial fit
        old_rms = []
        for ob in self.data.obs:
            tod = ob["tod"]
            orms = {}
            for det in tod.local_dets:
                cachename = "noise_{}".format(det)
                y = tod.cache.reference(cachename)
                x = np.arange(y.size)
                p = np.polyfit(x, y, self.order)
                y[:] = np.polyval(p, x)
                orms[det] = np.std(y)
                del y
            old_rms.append(orms)

        # Filter timestreams
        op = OpPolyFilter(name="noise", order=self.order)
        op.exec(self.data)

        # Ensure all timestreams are zeroed out by the filter.
        # The polynomial basis used in populating the TOD is different
        # than the one filter uses and the basis functions are not
        # strictly orthogonal on a sparse grid so we expect a low level
        # residual.

        for ob, orms in zip(self.data.obs, old_rms):
            tod = ob["tod"]
            for det in tod.local_dets:
                cachename = "noise_{}".format(det)
                y = tod.cache.reference(cachename)
                rms = np.std(y)
                old = orms[det]
                if np.abs(rms / old) > 1e-6:
                    raise RuntimeError(
                        "det {} old rms = {}, new rms = {}" "".format(det, old, rms)
                    )
                del y
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import sys
import os
import re
import traceback
import pickle

import numpy as np

available = True
try:
    from spt3g import core as c3g

    # from spt3g import coordinateutils as c3c
except ImportError:
    available = False


def from_g3_type(val):
    if isinstance(val, bool):
        return bool(val)
    elif isinstance(val, int):
        return int(val)
    elif isinstance(val, float):
        return float(val)
    else:
        check = str(val)
        if check == "NONE":
            return None
        else:
            return check
    return None


def g3_dtype(ar):
    if isinstance(ar, c3g.G3VectorUnsignedChar):
        return np.dtype(np.uint8)
    elif isinstance(ar, c3g.G3VectorInt):
        return np.dtype(np.int32)
    elif isinstance(ar, c3g.G3VectorTime):
        return np.dtype(np.int64)
    else:
        return np.dtype(np.float64)
    return None


def to_g3_type(val):
    if val is None:
        return c3g.G3String("NONE")
    elif isinstance(val, bool):
        return c3g.G3Bool(val)
    elif isinstance(val, int):
        return c3g.G3Int(val)
    elif isinstance(val, float):
        return c3g.G3Double(val)
    else:
        return c3g.G3String(val)


def read_spt3g_framesizes(file):
    sizes = list()
    for frame in c3g.G3File(file):
        if frame.type == c3g.G3FrameType.Scan:
            field = frame.keys()[0]
            sizes.append(len(frame[field]))
    return sizes


def compute_frame_offsets(frame_sizes):
    """Returns a list of the sample offsets for each frame.

    Args:
        frame_sizes (array-like): the size in samples of each frame.

    Returns:
        (array): the starting sample index of each frame.

    """
    ret = list()
    sampoff = 0
    for fr in frame_sizes:
        ret.append(sampoff)
        sampoff += fr
    return np.array(ret, dtype=np.int64)


def compute_file_frames(bytes_per_sample, frame_sizes, file_size=500000000):
    """Divide frames into approximately equal files.

    Given a list of frames and the number of bytes per sample of TOD, compute
    the file boundaries in terms of samples and frames.

    Args:
        bytes_per_sample (int): the number of bytes across all frame vectors
            for a single TOD sample.
        frame_sizes (array-like): the size in samples of each frame.
        file_size (int): the target file size in bytes.

    Returns:
        (tuple): the sample offsets and frame offsets of each file and the
            sample offsets of each frame.

    """
    sample_offs = list()
    frame_offs = list()
    frame_sample_offs = list()

    filebytes = 0
    filesamps = 0
    fileframes = 0
    fileoff = 0
    fileframeoff = 0
    sampoff = 0

    for fr in frame_sizes:
        frbytes = fr * bytes_per_sample
        if frbytes > file_size:
            msg = (
                "A single frame ({}) is larger than the target"
                " frame file size ({}). samples = {}, bytes/sample {}.  "
                "Increase the target"
                "size.".format(frbytes, file_size, fr, bytes_per_sample)
            )
            raise RuntimeError(msg)
        if filebytes + frbytes > file_size:
            # Start a new file
            sample_offs.append(fileoff)
            frame_offs.append(fileframeoff)
            fileoff += filesamps
            fileframeoff += fileframes
            filesamps = 0
            fileframes = 0
            filebytes = 0
        # Append frame to current file
        filesamps += fr
        fileframes += 1
        filebytes += frbytes
        frame_sample_offs.append(sampoff)
        sampoff += fr

    # process the last file
    sample_offs.append(fileoff)
    frame_offs.append(fileframeoff)

    return (
        np.array(sample_offs, dtype=np.int64),
        np.array(frame_offs, dtype=np.int64),
        np.array(frame_sample_offs, dtype=np.int64),
    )


def local_frame_indices(local_first, nlocal, frame_offset, frame_size):
    """Compute frame overlap with local data.

    Args:
        local_first (int): the first sample of the local data.
        nlocal (int): the number of local samples.
        frame_offset (int): the first sample of the frame.
        frame_size (int): the number of samples in the frame.

    Returns:
        tuple: the local sample offset (zero == start of local data) of the
            overlap data, the frame sample offset (zero == start of frame) of
            the overlap data, the number of samples in the overlap.

    """
    # The first sample in our local data buffer
    local_off = None

    # The first sample in the frame data buffer
    froff = None

    # The number of samples of overlap
    nsamp = None

    # Does this frame overlap with any of our data?
    if (frame_offset < local_first + nlocal) and (
        frame_offset + frame_size > local_first
    ):
        # compute offsets into our local data and the frame
        if frame_offset >= local_first:
            # The frame starts in the middle of our local sample range.
            local_off = frame_offset - local_first
            froff = 0
        else:
            # Our local samples start in the middle of the frame.
            local_off = 0
            froff = local_first - frame_offset
        nsamp = frame_size
        if local_off + nsamp > nlocal:
            # The frame extends beyond our local samples
            nsamp = nlocal - local_off

    return (local_off, froff, nsamp)


# FIXME:  The code below is not used in practice for any project, and more a up-to-date
# version of these functions specific to Simons Observatory is in:
#
#  https://github.com/simonsobs/sotodlib
#
# After identifying the general pieces of those tools, that code should be migrated here
# or to a new git repo focused on TOAST / SPT3G interfacing.

#
# def frame_to_cache(
#     tod,
#     frame,
#     frame_offset,
#     frame_size,
#     frame_data=None,
#     detector_map="detectors",
#     flag_map="flags",
#     common_prefix=None,
#     det_prefix=None,
#     flag_prefix=None,
# ):
#     """Distribute a frame from the rank zero process.
#
#     All vectors in the frame dictionary are copied into cache on all processes.
#     If the detector_map and / or flag_map options are specified, those are
#     frame keys which hold the timestream map and flag vector map.  When
#     distributing those, only keys in the maps that contain the local detector
#     names will be cached.
#
#     Args:
#         tod (toast.TOD): instance of a TOD class.
#         frame (int): the frame index.
#         frame_offset (int): the first sample of the frame.
#         frame_size (int): the number of samples in the the frame.
#         frame_data (G3Frame): the input frame (only on rank zero).
#         detector_map (str): the name of the frame timestream map.
#         flag_map (str): then name of the frame flag map.
#         common_prefix (str): a string to prepend to each field name when
#             storing in the cache.
#         det_prefix (str): a string to prepend to each field name when
#             storing in the cache.
#         flag_prefix (str): a string to prepend to each field name when
#             storing in the cache.
#
#     """
#     # First broadcast the frame data.
#     if tod.mpicomm is not None:
#         frame_data = tod.mpicomm.bcast(frame_data, root=0)
#
#     # Local sample range
#     local_first = tod.local_samples[0]
#     nlocal = tod.local_samples[1]
#
#     # Compute overlap of the frame with the local samples.
#     cacheoff, froff, nfr = local_frame_indices(
#         local_first, nlocal, frame_offset, frame_size
#     )
#
#     # Helper function to actually copy a slice of data into cache.
#     def copy_slice(data, fld, cache_prefix):
#         cache_fld = fld
#         if cache_prefix is not None:
#             cache_fld = "{}{}".format(cache_prefix, fld)
#         # Check field type and data shape
#         ftype = g3_dtype(data[fld])
#         flen = len(data[fld])
#         nnz = flen // frame_size
#         if nnz * frame_size != flen:
#             msg = (
#                 "frame {}, field {} has length {} which is not "
#                 "divisible by size {}".format(frame, fld, flen, frame_size)
#             )
#             raise RuntimeError(msg)
#         if not tod.cache.exists(cache_fld):
#             # The field does not yet exist in cache, so create it.
#             if nnz == 1:
#                 rf = tod.cache.create(cache_fld, ftype, (tod.local_samples[1],))
#             else:
#                 rf = tod.cache.create(cache_fld, ftype, (tod.local_samples[1], nnz))
#             del rf
#         rf = tod.cache.reference(cache_fld)
#         # Verify that the dimensions of the cache object are what we expect,
#         # then copy the data.
#         cache_samples = None
#         cache_nnz = None
#         if (len(rf.shape) > 1) and (rf.shape[1] > 0):
#             # We have a packed 2D array
#             cache_samples = rf.shape[0]
#             cache_nnz = rf.shape[1]
#         else:
#             cache_nnz = 1
#             cache_samples = len(rf)
#
#         if cache_samples != tod.local_samples[1]:
#             msg = "frame {}, field {}: cache has {} samples, which is"
#             " different from local TOD size {}".format(
#                 frame, fld, cache_samples, tod.local_samples[1]
#             )
#             raise RuntimeError(msg)
#
#         if cache_nnz != nnz:
#             msg = (
#                 "frame {}, field {}: cache has nnz = {}, which is"
#                 " different from frame nnz {}".format(frame, fld, cache_nnz, nnz)
#             )
#             raise RuntimeError(msg)
#
#         if cache_nnz > 1:
#             slc = np.array(
#                 data[fld][nnz * froff : nnz * (froff + nfr)], copy=False
#             ).reshape((-1, nnz))
#             rf[cacheoff : cacheoff + nfr, :] = slc
#         else:
#             slc = np.array(data[fld][froff : froff + nfr], copy=False)
#             rf[cacheoff : cacheoff + nfr] = slc
#         del rf
#         return
#
#     if cacheoff is not None:
#         # This process has some overlap with the frame.
#         for field in frame_data.keys():
#             # Skip over detector and flags
#             if (detector_map is not None) and (field == detector_map):
#                 continue
#             if (flag_map is not None) and (field == flag_map):
#                 continue
#             copy_slice(frame_data, field, common_prefix)
#
#         dpats = None
#         if (detector_map is not None) or (flag_map is not None):
#             # Build our list of regex matches
#             dpats = [re.compile(".*{}.*".format(d)) for d in tod.local_dets]
#
#         if detector_map is not None:
#             # If the field name contains any of our local detectors,
#             # then cache it.
#             for field in frame_data[detector_map].keys():
#                 for dp in dpats:
#                     if dp.match(field) is not None:
#                         copy_slice(frame_data[detector_map], field, det_prefix)
#                         break
#         if flag_map is not None:
#             # If the field name contains any of our local detectors,
#             # then cache it.
#             for field in frame_data[flag_map].keys():
#                 for dp in dpats:
#                     if dp.match(field) is not None:
#                         copy_slice(frame_data[flag_map], field, flag_prefix)
#                         break
#     return
#
#
# def cache_to_frames(
#     tod,
#     start_frame,
#     n_frames,
#     frame_offsets,
#     frame_sizes,
#     common=None,
#     detector_fields=None,
#     flag_fields=None,
#     detector_map="detectors",
#     flag_map="flags",
#     units=None,
# ):
#     """Gather all data from the distributed cache for a single frame.
#
#     Args:
#         tod (toast.TOD): instance of a TOD class.
#         start_frame (int): the first frame index.
#         n_frames (int): the number of frames.
#         frame_offsets (list): list of the first samples of all frames.
#         frame_sizes (list): list of the number of samples in each frame.
#         common (tuple): (cache name, G3 type, frame name) of each common
#             field.
#         detector_fields (tuple): (cache name, frame name) of each detector
#             field.
#         flag_fields (tuple): (cache name, frame name) of each flag field.
#         detector_map (str): the name of the frame timestream map.
#         flag_map (str): then name of the frame flag map.
#         units: G3 units of the detector data.
#
#     """
#     comm = tod.mpicomm
#     rank = 0
#     if comm is not None:
#         rank = comm.rank
#     comm_row = tod.grid_comm_row
#
#     # Local sample range
#     local_first = tod.local_samples[0]
#     nlocal = tod.local_samples[1]
#
#     # The process grid
#     detranks, sampranks = tod.grid_size
#     rankdet, ranksamp = tod.grid_ranks
#
#     # Helper function:
#     # For a given timestream, the gather is done across the
#     # process row which contains the specific detector, or across
#     # the first process row for common telescope data.
#     def gather_field(prow, fld, indx, cacheoff, ncache):
#         gproc = 0
#         gdata = None
#
#         # We are going to allreduce this later, so that every process
#         # knows the dimensions of the field.
#         allnnz = 0
#
#         if rankdet == prow:
#             # This process is in the process row that has this field,
#             # participate in the gather operation.
#             pdata = None
#             # Find the data type and shape from the cache object
#             mtype = None
#             ref = tod.cache.reference(fld)
#             nnz = 1
#             if (len(ref.shape) > 1) and (ref.shape[1] > 0):
#                 nnz = ref.shape[1]
#             if ref.dtype == np.dtype(np.float64):
#                 mtype = MPI.DOUBLE
#             elif ref.dtype == np.dtype(np.int64):
#                 mtype = MPI.INT64_T
#             elif ref.dtype == np.dtype(np.int32):
#                 mtype = MPI.INT32_T
#             elif ref.dtype == np.dtype(np.uint8):
#                 mtype = MPI.UINT8_T
#             else:
#                 msg = "Cannot gather cache field {} of type {}".format(fld, ref.dtype)
#                 raise RuntimeError(msg)
#             pz = 0
#             if cacheoff is not None:
#                 pdata = ref.flatten()[nnz * cacheoff : nnz * (cacheoff + ncache)]
#                 pz = nnz * ncache
#             else:
#                 pdata = np.zeros(0, dtype=ref.dtype)
#
#             psizes = tod.grid_comm_row.gather(pz, root=0)
#             disp = None
#             totsize = None
#             if ranksamp == 0:
#                 # print("Gathering field {} with type {}".format(fld, mtype), flush=True)
#                 # We are the process collecting the gathered data.
#                 gproc = tod.mpicomm.rank
#                 allnnz = nnz
#                 # Compute the displacements into the receive buffer.
#                 disp = [0]
#                 for ps in psizes[:-1]:
#                     last = disp[-1]
#                     disp.append(last + ps)
#                 totsize = np.sum(psizes)
#                 # allocate receive buffer
#                 gdata = np.zeros(totsize, dtype=ref.dtype)
#                 # print("Gatherv psizes = {}, disp = {}".format(psizes, disp), flush=True)
#
#             # print("field {}:  proc {} start Gatherv".format(fld, tod.mpicomm.rank), flush=True)
#             tod.grid_comm_row.Gatherv(pdata, [gdata, psizes, disp, mtype], root=0)
#             # print("field {}:  proc {} finish Gatherv".format(fld, tod.mpicomm.rank), flush=True)
#
#             del disp
#             del psizes
#             del pdata
#             del ref
#
#         # Now send this data to the root process of the whole communicator.
#         # Only one process (the first one in process row "prow") has data
#         # to send.
#
#         # Create a unique message tag
#         mtag = 10 * indx
#
#         # print("  proc {} hit allreduce of gproc".format(tod.mpicomm.rank), flush=True)
#         # All processes find out which one did the gather
#         gproc = tod.mpicomm.allreduce(gproc, MPI.SUM)
#         # All processes find out the field dimensions
#         allnnz = tod.mpicomm.allreduce(allnnz, MPI.SUM)
#         # print("  proc {} for field {}, gproc = {}".format(tod.mpicomm.rank, fld, gproc), flush=True)
#
#         # print("field {}:  proc {}, gatherproc = {}, allnnz = {}".format(fld, tod.mpicomm.rank, gproc, allnnz), flush=True)
#
#         rdata = None
#         if gproc == 0:
#             if gdata is not None:
#                 if allnnz == 1:
#                     rdata = gdata
#                 else:
#                     rdata = gdata.reshape((-1, allnnz))
#         else:
#             # Data not yet on rank 0
#             if tod.mpicomm.rank == 0:
#                 # Receive data from the first process in this row
#                 # print("  proc {} for field {}, recv type".format(tod.mpicomm.rank, fld), flush=True)
#                 rtype = tod.mpicomm.recv(source=gproc, tag=(mtag + 1))
#
#                 # print("  proc {} for field {}, recv size".format(tod.mpicomm.rank, fld), flush=True)
#                 rsize = tod.mpicomm.recv(source=gproc, tag=(mtag + 2))
#
#                 # print("  proc {} for field {}, recv data".format(tod.mpicomm.rank, fld), flush=True)
#                 rdata = np.zeros(rsize, dtype=np.dtype(rtype))
#                 tod.mpicomm.Recv(rdata, source=gproc, tag=mtag)
#
#                 # Reshape if needed
#                 if allnnz > 1:
#                     rdata = rdata.reshape((-1, allnnz))
#
#             elif tod.mpicomm.rank == gproc:
#                 # Send our data
#                 # print("  proc {} for field {}, send {} samples of {}".format(tod.mpicomm.rank, fld, len(gdata), gdata.dtype.char), flush=True)
#
#                 # print("  proc {} for field {}, send type with tag = {}".format(tod.mpicomm.rank, fld, mtag+1), flush=True)
#                 tod.mpicomm.send(gdata.dtype.char, dest=0, tag=(mtag + 1))
#
#                 # print("  proc {} for field {}, send size with tag = {}".format(tod.mpicomm.rank, fld, mtag+2), flush=True)
#                 tod.mpicomm.send(len(gdata), dest=0, tag=(mtag + 2))
#
#                 # print("  proc {} for field {}, send data with tag {}".format(tod.mpicomm.rank, fld, mtag), flush=True)
#                 tod.mpicomm.Send(gdata, 0, tag=mtag)
#         return rdata
#
#     # For efficiency, we are going to gather the data for all frames at once.
#     # Then we will split those up when doing the write.
#
#     # Frame offsets relative to the memory buffers we are gathering
#     fdataoff = [0]
#     for f in frame_sizes[:-1]:
#         last = fdataoff[-1]
#         fdataoff.append(last + f)
#
#     # The list of frames- only on the root process.
#     fdata = None
#     if tod.mpicomm.rank == 0:
#         fdata = [c3g.G3Frame(c3g.G3FrameType.Scan) for f in range(n_frames)]
#     else:
#         fdata = [None for f in range(n_frames)]
#
#     # Compute the overlap of all frames with the local process.  We want to
#     # to find the full sample range that this process overlaps the total set
#     # of frames.
#
#     cacheoff = None
#     ncache = 0
#
#     for f in range(n_frames):
#         # Compute overlap of the frame with the local samples.
#         fcacheoff, froff, nfr = local_frame_indices(
#             local_first, nlocal, frame_offsets[f], frame_sizes[f]
#         )
#         # print("proc {}:  frame {} has cache off {}, fr off {}, nfr {}".format(tod.mpicomm.rank, f, fcacheoff, froff, nfr), flush=True)
#         if fcacheoff is not None:
#             if cacheoff is None:
#                 cacheoff = fcacheoff
#                 ncache = nfr
#             else:
#                 ncache += nfr
#             # print("proc {}:    cache off now {}, ncache now {}".format(tod.mpicomm.rank, cacheoff, ncache), flush=True)
#
#     # Now gather the full sample data one field at a time.  The root process
#     # splits up the results into frames.
#
#     # First gather common fields from the first row of the process grid.
#
#     for findx, (cachefield, g3t, framefield) in enumerate(common):
#         # print("proc {} entering gather_field(0, {}, {}, {}, {})".format(tod.mpicomm.rank, cachefield, findx, cacheoff, ncache), flush=True)
#         data = gather_field(0, cachefield, findx, cacheoff, ncache)
#         if tod.mpicomm.rank == 0:
#             # print("Casting field {} to type {}".format(field, g3t), flush=True)
#             if g3t == c3g.G3VectorTime:
#                 # Special case for time values stored as int64_t, but
#                 # wrapped in a class.
#                 for f in range(n_frames):
#                     dataoff = fdataoff[f]
#                     ndata = frame_sizes[f]
#                     g3times = list()
#                     for t in range(ndata):
#                         g3times.append(c3g.G3Time(data[dataoff + t]))
#                     fdata[f][framefield] = c3g.G3VectorTime(g3times)
#                     del g3times
#             else:
#                 # The bindings of G3Vector seem to only work with
#                 # lists.  This is probably horribly inefficient.
#                 for f in range(n_frames):
#                     dataoff = fdataoff[f]
#                     ndata = frame_sizes[f]
#                     if len(data.shape) == 1:
#                         fdata[f][framefield] = g3t(
#                             data[dataoff : dataoff + ndata].tolist()
#                         )
#                     else:
#                         # We have a 2D quantity
#                         fdata[f][framefield] = g3t(
#                             data[dataoff : dataoff + ndata, :].flatten().tolist()
#                         )
#         del data
#
#     # Wait for everyone to catch up...
#     tod.mpicomm.barrier()
#
#     # For each detector field, processes which have the detector
#     # in their local_dets should be in the same process row.
#     # We do the gather over just this process row.
#
#     if (detector_fields is not None) or (flag_fields is not None):
#         dpats = {d: re.compile(".*{}.*".format(d)) for d in tod.local_dets}
#
#         detmaps = None
#         if detector_fields is not None:
#             if tod.mpicomm.rank == 0:
#                 detmaps = [c3g.G3TimestreamMap() for f in range(n_frames)]
#
#             for dindx, (cachefield, framefield) in enumerate(detector_fields):
#                 pc = -1
#                 for det, pat in dpats.items():
#                     if pat.match(cachefield) is not None:
#                         # print("proc {} has field {}".format(tod.mpicomm.rank, field), flush=True)
#                         pc = rankdet
#                         break
#                 # As a sanity check, verify that every process which
#                 # has this field is in the same process row.
#                 rowcheck = tod.mpicomm.gather(pc, root=0)
#                 prow = 0
#                 if tod.mpicomm.rank == 0:
#                     rc = np.array([x for x in rowcheck if (x >= 0)], dtype=np.int32)
#                     # print(field, rc, flush=True)
#                     prow = np.max(rc)
#                     if np.min(rc) != prow:
#                         msg = (
#                             "Processes with field {} are not in the "
#                             "same row\n".format(cachefield)
#                         )
#                         sys.stderr.write(msg)
#                         tod.mpicomm.abort()
#
#                 # Every process finds out which process row is participating.
#                 prow = tod.mpicomm.bcast(prow, root=0)
#                 # print("proc {} got prow = {}".format(tod.mpicomm.rank, prow), flush=True)
#
#                 # Get the data on rank 0
#                 data = gather_field(prow, cachefield, dindx, cacheoff, ncache)
#
#                 if tod.mpicomm.rank == 0:
#                     if units is None:
#                         # We do this conditional, since we can't use
#                         # G3TimestreamUnits.None in python ("None" is
#                         # interpreted as python None).
#                         for f in range(n_frames):
#                             dataoff = fdataoff[f]
#                             ndata = frame_sizes[f]
#                             detmaps[f][framefield] = c3g.G3Timestream(
#                                 data[dataoff : dataoff + ndata]
#                             )
#                     else:
#                         for f in range(n_frames):
#                             dataoff = fdataoff[f]
#                             ndata = frame_sizes[f]
#                             detmaps[f][framefield] = c3g.G3Timestream(
#                                 data[dataoff : dataoff + ndata], units
#                             )
#
#             if tod.mpicomm.rank == 0:
#                 for f in range(n_frames):
#                     fdata[f][detector_map] = detmaps[f]
#
#         flagmaps = None
#         if flag_fields is not None:
#             if tod.mpicomm.rank == 0:
#                 flagmaps = [c3g.G3MapVectorInt() for f in range(n_frames)]
#             for dindx, (cachefield, framefield) in enumerate(flag_fields):
#                 pc = -1
#                 for det, pat in dpats.items():
#                     if pat.match(cachefield) is not None:
#                         pc = rankdet
#                         break
#                 # As a sanity check, verify that every process which
#                 # has this field is in the same process row.
#                 rowcheck = tod.mpicomm.gather(pc, root=0)
#                 prow = 0
#                 if tod.mpicomm.rank == 0:
#                     rc = np.array([x for x in rowcheck if (x >= 0)], dtype=np.int32)
#                     prow = np.max(rc)
#                     if np.min(rc) != prow:
#                         msg = (
#                             "Processes with field {} are not in the "
#                             "same row\n".format(cachefield)
#                         )
#                         sys.stderr.write(msg)
#                         tod.mpicomm.abort()
#
#                 # Every process finds out which process row is participating.
#                 prow = tod.mpicomm.bcast(prow, root=0)
#
#                 # Get the data on rank 0
#                 data = gather_field(prow, cachefield, dindx, cacheoff, ncache)
#
#                 if tod.mpicomm.rank == 0:
#                     # The bindings of G3Vector seem to only work with
#                     # lists...  Also there is no vectormap for unsigned
#                     # char, so we have to use int...
#                     for f in range(n_frames):
#                         dataoff = fdataoff[f]
#                         ndata = frame_sizes[f]
#                         flagmaps[f][framefield] = c3g.G3VectorInt(
#                             data[dataoff : dataoff + ndata].astype(np.int32).tolist()
#                         )
#
#             if tod.mpicomm.rank == 0:
#                 for f in range(n_frames):
#                     fdata[f][flag_map] = flagmaps[f]
#
#     return fdata



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

import healpy as hp

from ..qarray import rotate

from ..utils import Environment

from ..healpix import HealpixPixels

from ..op import Operator

from ..timing import function_timer

from .._libtoast import pointing_matrix_healpix


class OpPointingHpix(Operator):
    """
    Operator which generates I/Q/U healpix pointing weights.

    Given the individual detector pointing, this computes the pointing weights
    assuming that the detector is a linear polarizer followed by a total
    power measurement.  An optional dictionary of calibration factors may
    be specified.  Additional options include specifying a constant cross-polar
    response (eps) and a rotating, perfect half-wave plate.  The timestream
    model is then (see Jones, et al, 2006):

    .. math::
        d = cal \\left[\\frac{(1+eps)}{2} I + \\frac{(1-eps)}{2} \\left[Q \\cos{2a} + U \\sin{2a}\\right]\\right]

    Or, if a HWP is included in the response with time varying angle "w", then
    the total response is:

    .. math::
        d = cal \\left[\\frac{(1+eps)}{2} I + \\frac{(1-eps)}{2} \\left[Q \\cos{2a+4w} + U \\sin{2a+4w}\\right]\\right]

    Args:
        pixels (str): write pixels to the cache with name <pixels>_<detector>.
            If the named cache objects do not exist, then they are created.
        weights (str): write pixel weights to the cache with name
            <weights>_<detector>.  If the named cache objects do not exist,
            then they are created.
        nside (int): NSIDE resolution for Healpix NEST ordered intensity map.
        nest (bool): if True, use NESTED ordering.
        mode (string): either "I" or "IQU"
        cal (dict): dictionary of calibration values per detector. A None
            value means a value of 1.0 for all detectors.
        epsilon (dict): dictionary of cross-polar response per detector. A
            None value means epsilon is zero for all detectors.
        common_flag_name (str): the optional name of a cache object to use for
            the common flags.
        common_flag_mask (byte): the bitmask to use when flagging the pointing
            matrix using the common flags.
        apply_flags (bool): whether to read the TOD common flags, bitwise OR
            with the common_flag_mask, and then flag the pointing matrix.
        single_precision (bool):  Return the pixel numbers and pointing
             weights in single precision.  Default=False.
        nside_submap (int):  Size of a submap is 12 * nside_submap ** 2
    """

    def __init__(
        self,
        pixels="pixels",
        weights="weights",
        nside=64,
        nest=False,
        mode="I",
        cal=None,
        epsilon=None,
        common_flag_name=None,
        common_flag_mask=255,
        apply_flags=False,
        keep_quats=False,
        single_precision=False,
        nside_submap=16,
    ):
        self._pixels = pixels
        self._weights = weights
        self._nside = nside
        self._nest = nest
        self._mode = mode
        self._cal = cal
        self._epsilon = epsilon
        self._common_flag_mask = common_flag_mask
        self._apply_flags = apply_flags
        self._common_flag_name = common_flag_name
        self._keep_quats = keep_quats
        self._single_precision = single_precision
        self._nside_submap = min(nside, nside_submap)
        self._npix_submap = 12 * self._nside_submap ** 2
        self._nsubmap = (self._nside // self._nside_submap) ** 2
        self._hit_submaps = np.zeros(self._nsubmap, dtype=np.bool)

        # initialize the healpix pixels object
        self.hpix = HealpixPixels(self._nside)

        if self._mode == "I":
            self._nnz = 1
        elif self._mode == "IQU":
            self._nnz = 3
        else:
            raise RuntimeError("Unsupported mode")

        # We call the parent class constructor, which currently does nothing
        super().__init__()

    @property
    def nside(self):
        """(int): the HEALPix NSIDE value used."""
        return self._nside

    @property
    def nest(self):
        """(bool): if True, the pointing is NESTED ordering."""
        return self._nest

    @property
    def mode(self):
        """(str): the pointing mode "I", "IQU", etc."""
        return self._mode

    @property
    def local_submaps(self):
        """(list): Indices of locally hit submaps"""
        if self._single_precision:
            dtype = np.int32
        else:
            dtype = np.int64
        return np.arange(self._nsubmap, dtype=dtype)[self._hit_submaps]

    @function_timer
    def exec(self, data):
        """Create pixels and weights.

        This iterates over all observations and detectors, and creates
        the pixel and weight arrays representing the pointing matrix.
        This data is stored in the TOD cache.

        Args:
            data (toast.Data): The distributed data.

        """
        env = Environment.get()
        tod_buffer_length = env.tod_buffer_length()

        for obs in data.obs:
            tod = obs["tod"]

            # compute effective sample rate

            times = tod.local_times()
            dt = np.mean(times[1:-1] - times[0:-2])
            rate = 1.0 / dt
            del times

            offset, nsamp = tod.local_samples

            hwpang = None
            try:
                hwpang = tod.local_hwp_angle()
            except:
                hwpang = None

            # read the common flags and apply bitmask

            common = None
            if self._apply_flags:
                common = tod.local_common_flags(self._common_flag_name)
                common = common & self._common_flag_mask
            else:
                common = np.zeros(nsamp, dtype=np.uint8)

            for det in tod.local_dets:
                eps = 0.0
                if self._epsilon is not None:
                    eps = self._epsilon[det]

                cal = 1.0
                if self._cal is not None:
                    cal = self._cal[det]

                # Create cache objects and use that memory directly

                pixelsname = "{}_{}".format(self._pixels, det)
                weightsname = "{}_{}".format(self._weights, det)

                pixelsref = None
                weightsref = None

                if tod.cache.exists(pixelsname):
                    pixelsref = tod.cache.reference(pixelsname)
                else:
                    pixelsref = tod.cache.create(pixelsname, np.int64, (nsamp,))

                if tod.cache.exists(weightsname):
                    weightsref = tod.cache.reference(weightsname)
                else:
                    weightsref = tod.cache.create(
                        weightsname, np.float64, (nsamp, self._nnz)
                    )

                pdata = None
                if self._keep_quats:
                    # We are keeping the detector quaternions, so cache
                    # them now for the full sample range.
                    pdata = tod.local_pointing(det)

                buf_off = 0
                buf_n = tod_buffer_length
                while buf_off < nsamp:
                    if buf_off + buf_n > nsamp:
                        buf_n = nsamp - buf_off
                    bslice = slice(buf_off, buf_off + buf_n)

                    detp = None
                    if pdata is None:
                        # Read and discard
                        detp = tod.read_pntg(detector=det, local_start=buf_off, n=buf_n)
                    else:
                        # Use cached version
                        detp = pdata[bslice, :]

                    hslice = None
                    if hwpang is not None:
                        hslice = hwpang[bslice].reshape(-1)
                    fslice = common[bslice].reshape(-1)

                    pointing_matrix_healpix(
                        self.hpix,
                        self._nest,
                        eps,
                        cal,
                        self._mode,
                        detp.reshape(-1),
                        hslice,
                        fslice,
                        pixelsref[bslice].reshape(-1),
                        weightsref[bslice, :].reshape(-1),
                    )
                    buf_off += buf_n

                if self._single_precision:
                    pixels = pixelsref.astype(np.int32)
                    del pixelsref
                    pixelsref = tod.cache.put(pixelsname, pixels, replace=True)
                    del pixels
                    weights = weightsref.astype(np.float32)
                    del weightsref
                    weightsref = tod.cache.put(weightsname, weights, replace=True)
                    del weights

                self._hit_submaps[pixelsref // self._npix_submap] = True

                del pixelsref
                del weightsref
                del pdata

            del common

        # Store the local submaps in the data object under the same name
        # as the pixel numbers

        if self._single_precision:
            dtype = np.int32
        else:
            dtype = np.int64

        local_submaps = np.arange(self._nsubmap, dtype=dtype)[self._hit_submaps]
        submap_name = "{}_local_submaps".format(self._pixels)
        data[submap_name] = local_submaps
        npix_submap_name = "{}_npix_submap".format(self._pixels)
        data[npix_submap_name] = self._npix_submap
        nsubmap_name = "{}_nsubmap".format(self._pixels)
        data[nsubmap_name] = self._nsubmap
        npix_name = "{}_npix".format(self._pixels)
        data[npix_name] = 12 * self._nside ** 2

        return


class OpMuellerPointingHpix(Operator):
    """Operator which generates I/Q/U healpix pointing weights.

    Given the individual detector pointing, this computes the pointing weights
    assuming that the detector is a linear polarizer followed by a total
    power measurement.  An optional dictionary of calibration factors may
    be specified.  Additional options include specifying a constant cross-polar
    response (eps) and a set of HWP Mueller matrix parameters.  The timestream
    model is then (see Jones, et al, 2006):

    .. math::
        d = cal \\left[\\frac{(1+eps)}{2} I + \\frac{(1-eps)}{2} \\left[Q \\cos{2a} + U \\sin{2a}\\right]\\right]

    Or, if a HWP is included in the response with time varying angle "w", then
    the total response is:

    .. math::
        d = cal \\left[\\frac{(1+eps)}{2} I + \\frac{(1-eps)}{2} \\left[Q \\cos{2a+4w} + U \\sin{2a+4w}\\right]\\right]

    Args:
        pixels (str): write pixels to the cache with name <pixels>_<detector>.
            If the named cache objects do not exist, then they are created.
        weights (str): write pixel weights to the cache with name
            <weights>_<detector>.  If the named cache objects do not exist,
            then they are created.
        nside (int): NSIDE resolution for Healpix NEST ordered intensity map.
        nest (bool): if True, use NESTED ordering.
        mode (string): either "I" or "IQU"
        cal (dict): dictionary of calibration values per detector. A None
            value means a value of 1.0 for all detectors.
        epsilon (dict): dictionary of cross-polar response per detector. A
            None value means epsilon is zero for all detectors.
        common_flag_name (str): the optional name of a cache object to use for
            the common flags.
        common_flag_mask (byte): the bitmask to use when flagging the pointing
            matrix using the common flags.
        apply_flags (bool): whether to read the TOD common flags, bitwise OR
            with the common_flag_mask, and then flag the pointing matrix.
        single_precision (bool):  Return the pixel numbers and pointing
            weights in single precision.  Default=False.
        nside_submap (int):  Size of a submap is 12 * nside_submap ** 2
        hwp_parameters_set (tuple) : tuple of  parameters describing
            the Mueller Matrix [T, c,rho,s ] as defined in Bryan et al. 2010 :

            .. math::
                M_{HWP}=
                \begin{pmatrix}
                    T & \rho  & 0 & 0 \\
                    \rho & T & 0 &  0 \\
                      0  & 0  & c & -s  \\
                      0  & 0  & s & c  \\
                \end{pmatrix}

            default assumes an ideal HWP i.e. T=1,c=-1,rho=0,s=0.

    """

    def __init__(
        self,
        pixels="pixels",
        weights="weights",
        nside=64,
        nest=False,
        mode="I",
        cal=None,
        epsilon=None,
        common_flag_name=None,
        common_flag_mask=255,
        apply_flags=False,
        keep_quats=False,  # this should be set to True to work in L480
        single_precision=False,
        nside_submap=16,
        hwp_parameters_set=[1, -1, 0, 0],
    ):
        self._pixels = pixels
        self._weights = weights
        self._nside = nside
        self._nest = nest
        self._mode = mode
        self._cal = cal
        self._epsilon = epsilon
        self._common_flag_mask = common_flag_mask
        self._apply_flags = apply_flags
        self._common_flag_name = common_flag_name
        self._keep_quats = keep_quats
        self._single_precision = single_precision
        self._nside_submap = min(nside, nside_submap)
        self._npix_submap = 12 * self._nside_submap ** 2
        self._nsubmap = (self._nside // self._nside_submap) ** 2
        self._hit_submaps = np.zeros(self._nsubmap, dtype=np.bool)
        self._hwp_parameters_set = hwp_parameters_set
        # initialize the healpix pixels object
        self.hpix = HealpixPixels(self._nside)

        if self._mode == "I":
            self._nnz = 1
        elif self._mode == "IQU":
            self._nnz = 3
        else:
            raise RuntimeError("Unsupported mode")

        # We call the parent class constructor, which currently does nothing
        super().__init__()

    @property
    def nside(self):
        """(int): the HEALPix NSIDE value used."""
        return self._nside

    @property
    def nest(self):
        """(bool): if True, the pointing is NESTED ordering."""
        return self._nest

    @property
    def mode(self):
        """(str): the pointing mode "I", "IQU", etc."""
        return self._mode

    @property
    def local_submaps(self):
        """(list): Indices of locally hit submaps"""
        if self._single_precision:
            dtype = np.int32
        else:
            dtype = np.int64
        return np.arange(self._nsubmap, dtype=dtype)[self._hit_submaps]

    @function_timer
    def exec(self, data):
        """Create pixels and weights.

        This iterates over all observations and detectors, and creates
        the pixel and weight arrays representing the pointing matrix.
        This data is stored in the TOD cache.

        Args:
            data (toast.Data): The distributed data.

        """
        env = Environment.get()
        tod_buffer_length = env.tod_buffer_length()

        for obs in data.obs:
            tod = obs["tod"]

            # compute effective sample rate

            times = tod.local_times()
            dt = np.mean(times[1:-1] - times[0:-2])
            rate = 1.0 / dt
            del times

            offset, nsamp = tod.local_samples

            hwpang = None
            try:
                hwpang = tod.local_hwp_angle()
            except:
                hwpang = None
                raise RuntimeError("Can't run this operator if hwpang is None ")

            # read the common flags and apply bitmask

            common = None
            if self._apply_flags:
                common = tod.local_common_flags(self._common_flag_name)
                common = common & self._common_flag_mask
            else:
                common = np.zeros(nsamp, dtype=np.uint8)

            for det in tod.local_dets:
                eps = 0.0
                if self._epsilon is not None:
                    eps = self._epsilon[det]

                cal = 1.0
                if self._cal is not None:
                    cal = self._cal[det]

                # Create cache objects and use that memory directly

                pixelsname = "{}_{}".format(self._pixels, det)
                weightsname = "{}_{}".format(self._weights, det)

                pixelsref = None
                weightsref = None

                if tod.cache.exists(pixelsname):
                    pixelsref = tod.cache.reference(pixelsname)
                else:
                    pixelsref = tod.cache.create(pixelsname, np.int64, (nsamp,))

                if tod.cache.exists(weightsname):
                    weightsref = tod.cache.reference(weightsname)
                else:
                    weightsref = tod.cache.create(
                        weightsname, np.float64, (nsamp, self._nnz)
                    )

                pdata = None
                if self._keep_quats:
                    # We are keeping the detector quaternions, so cache
                    # them now for the full sample range.
                    pdata = tod.local_pointing(det)

                xaxis = np.array([1.0, 0.0, 0.0])
                zaxis = np.array([0.0, 0.0, 1.0])
                nullquat = np.array([0.0, 0.0, 0.0, 1.0])
                eta = (1.0 - eps) / (1.0 + eps)

                pdata[common] = nullquat

                dir = rotate(pdata, zaxis)
                pixelsref = hp.vec2pix(
                    nside=self._nside, x=dir[:, 0], y=dir[:, 1], z=dir[:, 2], nest=True
                )

                pixelsref[common] = -1
                # import pdb
                # pdb.set_trace()

                T = self._hwp_parameters_set[0]
                c = self._hwp_parameters_set[1]
                rho = self._hwp_parameters_set[2]
                s = self._hwp_parameters_set[3]
                cos2hwp = np.cos(2 * hwpang)

                if self._mode == "I":
                    weightsref = cal * (T + eta * rho * cos2hwp)
                elif self._mode == "IQU":
                    orient = rotate(pdata, xaxis)

                    by = orient[:, 0] * dir[:, 1] - orient[:, 1] * dir[:, 0]
                    bx = (
                        orient[:, 0] * (-dir[:, 2] * dir[:, 0])
                        + orient[:, 1] * (-dir[:, 2] * dir[:, 1])
                        + orient[:, 2] * (dir[:, 0] * dir[:, 0] + dir[:, 1] * dir[:, 1])
                    )

                    detang = np.arctan2(by, bx)

                    sindetang = np.sin(2 * detang)
                    cosdetang = np.cos(2 * detang)

                    ang4 = 2 * detang + 4 * hwpang
                    ang2 = 2 * detang + 2 * hwpang

                    sin4 = np.sin(ang4)
                    sin2 = np.sin(ang2)
                    cos4 = np.cos(ang4)
                    cos2 = np.cos(ang2)

                    weightsref[:, 0] = cal * (T + (eta * rho * cos2hwp))
                    weightsref[:, 1] = cal * (
                        (rho * cos2)
                        + eta * (T + c) / 2.0 * cosdetang
                        + eta * (T - c) / 2.0 * cos4
                    )
                    weightsref[:, 2] = cal * (
                        (rho * sin2)
                        + eta * (T + c) / 2.0 * sindetang
                        + eta * (T - c) / 2.0 * sin4
                    )
                else:
                    raise RuntimeError("Unknown healpix pointing matrix mode")

                if self._single_precision:
                    pixels = pixelsref.astype(np.int32)
                    del pixelsref
                    pixelsref = tod.cache.put(pixelsname, pixels, replace=True)
                    del pixels
                    weights = weightsref.astype(np.float32)
                    del weightsref
                    weightsref = tod.cache.put(weightsname, weights, replace=True)
                    del weights

                self._hit_submaps[pixelsref // self._npix_submap] = True

                del pixelsref
                del weightsref
                del pdata

            del common

        # Store the local submaps in the data object under the same name
        # as the pixel numbers

        if self._single_precision:
            dtype = np.int32
        else:
            dtype = np.int64

        local_submaps = np.arange(self._nsubmap, dtype=dtype)[self._hit_submaps]
        submap_name = "{}_local_submaps".format(self._pixels)
        data[submap_name] = local_submaps
        npix_submap_name = "{}_npix_submap".format(self._pixels)
        data[npix_submap_name] = self._npix_submap
        nsubmap_name = "{}_nsubmap".format(self._pixels)
        data[nsubmap_name] = self._nsubmap
        npix_name = "{}_npix".format(self._pixels)
        data[npix_name] = 12 * self._nside ** 2

        return



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import os

import numpy as np

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

from ..tod import tidas_available, spt3g_available

if tidas_available:
    from ..tod.tidas import OpTidasExport, TODTidas

# if spt3g_available:
#    from ..tod.spt3g import Op3GExport, TOD3G


def add_tidas_args(parser):
    """Add the noise simulation arguments"""
    parser.add_argument(
        "--tidas", required=False, default=None, help="Output TIDAS export path"
    )


def add_spt3g_args(parser):
    """Add the noise simulation arguments"""
    parser.add_argument(
        "--spt3g", required=False, default=None, help="Output SPT3G export path"
    )


@function_timer
def output_tidas(args, comm, data, cache_prefix=None, verbose=True):
    if args.tidas is None:
        return
    if not tidas_available:
        raise RuntimeError(
            "TIDAS not available.  Cannot export to '{}'" "".format(args.tidas)
        )
    log = Logger.get()
    timer = Timer()
    tidas_path = os.path.abspath(args.tidas)

    if comm.world_rank == 0 and verbose:
        log.info("Exporting data to a TIDAS volume at {}".format(tidas_path))

    timer.start()
    export = OpTidasExport(
        tidas_path,
        TODTidas,
        backend="hdf5",
        use_intervals=True,
        create_opts={"group_dets": "sim"},
        ctor_opts={"group_dets": "sim"},
        cache_name=cache_prefix,
    )
    export.exec(data)

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    timer.stop()
    if comm.world_rank == 0 and verbose:
        timer.report("TIDAS export")


@function_timer
def output_spt3g(args, comm, data, cache_prefix=None, verbose=True):
    """
    if args.spt3g is None:
        return
    if not spt3g_available:
        raise RuntimeError(
            "SPT3G not available.  Cannot export to '{}'" "".format(args.spt3g)
        )
    log = Logger.get()
    timer = Timer()

    spt3g_path = os.path.abspath(args.spt3g)

    if comm.world_rank == 0 and verbose:
        log.info("Exporting data to a SPT3G directory tree at {}" "".format(spt3g_path))

    timer.start()
    export = Op3GExport(
        spt3g_path,
        TOD3G,
        use_intervals=True,
        export_opts={"prefix": "sim"},
        cache_name=cache_prefix,
    )
    export.exec(data)

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    timer.stop()
    if comm.world_rank == 0 and verbose:
        timer.report("spt3g export")
    """



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import dateutil.parser
import os

import healpy as hp
import numpy as np

from .. import qarray
from ..timing import function_timer, Timer
from ..utils import Logger, Environment
from ..weather import Weather

from .classes import Telescope, Focalplane
from .debug import add_debug_args

# Schedule, Site and CES are small helper classes for building
# ground observations


class Schedule:
    def __init__(self, telescope=None, ceslist=None, sort=False):
        self.telescope = telescope
        self.ceslist = ceslist
        if sort:
            self.sort_ceslist()
        return

    def sort_ceslist(self):
        """Sort the list of CES by name"""
        nces = len(self.ceslist)
        for i in range(nces - 1):
            for j in range(i + 1, nces):
                if self.ceslist[j].name < self.ceslist[j - 1].name:
                    temp = self.ceslist[j]
                    self.ceslist[j] = self.ceslist[j - 1]
                    self.ceslist[j - 1] = temp
        return


class Site:
    def __init__(self, name, lat, lon, alt, weather=None):
        """Instantiate a Site object

        args:
            name (str)
            lat (str) :  Site latitude as a pyEphem string
            lon (str) :  Site longitude as a pyEphem string
            alt (float) :  Site altitude in meters
            telescope (str) :  Optional telescope instance at the site
        """
        self.name = name
        # Strings get interpreted correctly pyEphem.
        # Floats must be in radians
        self.lat = str(lat)
        self.lon = str(lon)
        self.alt = alt
        self.id = 0
        self.weather = weather

    def __repr__(self):
        value = (
            "(Site '{}' : ID = {}, lon = {}, lat = {}, alt = {} m, "
            "weather = {})"
            "".format(self.name, self.id, self.lon, self.lat, self.alt, self.weather)
        )
        return value


class CES:
    def __init__(
        self,
        start_time,
        stop_time,
        boresight_angle,
        name,
        mjdstart,
        scan,
        subscan,
        azmin,
        azmax,
        el,
        season,
        start_date,
        rising,
        mindist_sun,
        mindist_moon,
        el_sun,
    ):
        self.start_time = start_time
        self.stop_time = stop_time
        self.boresight_angle = boresight_angle
        self.name = name
        self.mjdstart = mjdstart
        self.scan = scan
        self.subscan = subscan
        self.azmin = azmin
        self.azmax = azmax
        self.el = el
        self.season = season
        self.start_date = start_date
        self.rising = rising
        self.mindist_sun = mindist_sun
        self.mindist_moon = mindist_moon
        self.el_sun = el_sun


def add_todground_args(parser):
    """Add TODGround arguments"""
    parser.add_argument(
        "--el-mod-step-deg",
        required=False,
        default=0,
        type=np.float,
        help="If non-zero, scanning elevation will be stepped "
        "after each scan pair (upon returning to starting "
        "azimuth). [degrees]",
    )
    parser.add_argument(
        "--el-mod-rate-hz",
        required=False,
        default=0,
        type=np.float,
        help="If non-zero, observing elevation will be "
        "continuously modulated during the science scan. [Hz]",
    )
    parser.add_argument(
        "--el-mod-amplitude-deg",
        required=False,
        default=1,
        type=np.float,
        help="Range of elevation modulation when "
        "`el_mod_rate` is non-zero. [degrees]",
    )
    parser.add_argument(
        "--el-mod-sine",
        required=False,
        action="store_true",
        help="Elevation modulation should produce a sine pattern",
        dest="el_mod_sine",
    )
    parser.add_argument(
        "--no-el-mod-sine",
        required=False,
        action="store_false",
        help="Elevation modulation should produce a triangle wave",
        dest="el_mod_sine",
    )
    parser.set_defaults(el_mod_sine=False)
    parser.add_argument(
        "--el-nod-deg",
        required=False,
        help="Comma-separated list of elevation changes to perform in degrees",
    )
    parser.add_argument(
        "--el-nod-every-scan",
        required=False,
        action="store_true",
        help="Append el nods to every scan, not just the first one",
        dest="el_nod_every_scan",
    )
    parser.add_argument(
        "--no-el-nod-every-scan",
        required=False,
        action="store_false",
        help="Only append el-nod to the first sub scan",
        dest="el_nod_every_scan",
    )
    parser.set_defaults(el_nod_every_scan=True)
    parser.add_argument(
        "--start-with-el-nod",
        required=False,
        action="store_true",
        help="When requested, el-nod should be appended to the beginning of each CES",
        dest="start_with_el_nod",
    )
    parser.add_argument(
        "--no-start-with-el-nod",
        required=False,
        action="store_false",
        help="When requested, el-nod should not be appended to the beginning of each CES",
        dest="start_with_el_nod",
    )
    parser.set_defaults(start_with_el_nod=True)
    parser.add_argument(
        "--end-with-el-nod",
        required=False,
        action="store_true",
        help="When requested, el-nod should be appended to the end of each CES",
        dest="end_with_el_nod",
    )
    parser.add_argument(
        "--no-end-with-el-nod",
        required=False,
        action="store_false",
        help="When requested, el-nod should not be appended to the end of each CES",
        dest="end_with_el_nod",
    )
    parser.set_defaults(end_with_el_nod=False)
    parser.add_argument(
        "--scan-rate",
        required=False,
        default=1.0,
        type=np.float,
        help="Azimuthal on-sky scanning rate [deg / s]",
    )
    parser.add_argument(
        "--scan-rate-el",
        required=False,
        type=np.float,
        help="Elevation scanning rate [deg / s]",
    )
    parser.add_argument(
        "--scan-accel",
        required=False,
        default=1.0,
        type=np.float,
        help="Mount azimuthal scanning rate change [deg / s^2]",
    )
    parser.add_argument(
        "--scan-accel-el",
        required=False,
        type=np.float,
        help="Mount elevation scanning rate change [deg / s^2]",
    )
    parser.add_argument(
        "--scan-cosecant-modulate",
        required=False,
        action="store_true",
        help="Modulate scan rate so integration depth is constant at "
        "all declinations.  The --scan-rate becomes the *minimum* scan rate.",
        dest="scan_cosecant_modulate",
    )
    parser.add_argument(
        "--no-scan-cosecant-modulate",
        required=False,
        action="store_false",
        help="Use constant sky scan rate according to --scan-rate.",
        dest="scan_cosecant_modulate",
    )
    parser.set_defaults(scan_cosecant_modulate=False)
    parser.add_argument(
        "--sun-angle-min",
        required=False,
        default=30.0,
        type=np.float,
        help="Minimum azimuthal distance between the Sun and the bore sight [deg]",
    )
    parser.add_argument(
        "--schedule",
        required=True,
        help="Comma-separated list CES schedule files "
        "(from toast_ground_schedule.py)",
    )
    parser.add_argument(
        "--weather",
        required=False,
        help="Comma-separated list of TOAST weather files for "
        "every schedule.  Repeat the same file if the "
        "schedules share observing site.",
    )
    parser.add_argument(
        "--timezone",
        required=False,
        type=np.int,
        default=0,
        help="Offset to apply to MJD to separate days [hours]",
    )

    add_debug_args(parser)

    # `sample-rate` may be already added
    try:
        parser.add_argument(
            "--sample-rate",
            required=False,
            default=100.0,
            type=np.float,
            help="Detector sample rate (Hz)",
        )
    except argparse.ArgumentError:
        pass
    # `coord` may already be added
    try:
        parser.add_argument(
            "--coord", required=False, default="C", help="Sky coordinate system [C,E,G]"
        )
    except argparse.ArgumentError:
        pass
    parser.add_argument(
        "--split-schedule",
        required=False,
        help="Only use a subset of the schedule.  The argument is a string "
        'of the form "[isplit],[nsplit]" and only observations that satisfy '
        "scan modulo nsplit == isplit are included",
    )
    parser.add_argument(
        "--sort-schedule",
        required=False,
        action="store_true",
        help="Reorder the observing schedule so that observations of the "
        "same patch are consecutive.  This will reduce the sky area observed "
        "by individual process groups.",
        dest="sort_schedule",
    )
    parser.add_argument(
        "--no-sort-schedule",
        required=False,
        action="store_false",
        help="Do not reorder the observing schedule so that observations of the "
        "same patch are consecutive.",
        dest="sort_schedule",
    )
    parser.set_defaults(sort_schedule=True)

    # The HWP arguments may also be added by other TOD classes
    try:
        parser.add_argument(
            "--hwp-rpm",
            required=False,
            type=np.float,
            help="The rate (in RPM) of the HWP rotation",
        )
    except argparse.ArgumentError:
        pass
    try:
        parser.add_argument(
            "--hwp-step-deg",
            required=False,
            type=np.float,
            help="For stepped HWP, the angle in degrees of each step",
        )
    except argparse.ArgumentError:
        pass
    try:
        parser.add_argument(
            "--hwp-step-time-s",
            required=False,
            type=np.float,
            help="For stepped HWP, the time in seconds between steps",
        )
    except argparse.ArgumentError:
        pass
    # Modulate noise PSD by observing elevation
    parser.add_argument(
        "--elevation-noise-a",
        default=0,
        type=np.float,
        help="Evaluate noise PSD as (a / sin(el) + b) ** 2 * fsample * 1e-12",
    )
    parser.add_argument(
        "--elevation-noise-b",
        default=0,
        type=np.float,
        help="Evaluate noise PSD as (a / sin(el) + b) ** 2 * fsample * 1e-12",
    )

    return


@function_timer
def get_elevation_noise(args, comm, data, key="noise"):
    """Insert elevation-dependent noise"""
    if args.elevation_noise_a == 0 and args.elevation_noise_b == 0:
        return
    timer = Timer()
    timer.start()
    a = args.elevation_noise_a
    b = args.elevation_noise_b
    fsample = args.sample_rate
    for obs in data.obs:
        tod = obs["tod"]
        noise = obs[key]
        for det in tod.local_dets:
            if det not in noise.keys:
                raise RuntimeError(
                    'Detector "{}" does not have a PSD in the noise object'.format(det)
                )
            # freq = noise.freq[det]
            psd = noise.psd(det)
            try:
                # Some TOD classes provide a shortcut to Az/El
                _, el = tod.read_azel(detector=det)
            except Exception as e:
                nlocal = tod.local_samples[1]
                local_start = nlocal // 2 - nlocal // 20
                n = nlocal // 10
                azelquat = tod.read_pntg(
                    detector=det, local_start=local_start, n=n, azel=True
                )
                # Convert Az/El quaternion of the detector back into
                # angles for the simulation.
                theta, _ = qarray.to_position(azelquat)
                el = np.pi / 2 - theta
            # The model evaluates to uK / sqrt(Hz)
            # Translate it to K_CMB ** 2
            el = np.median(el)
            old_net = np.median(psd[-10:])
            new_net = (a / np.sin(el) + b) ** 2 * fsample * 1e-12
            psd[:] *= new_net / old_net
    if comm.comm_world is None or comm.world_rank == 0:
        timer.report_clear("Elevation noise")
    return


@function_timer
def get_breaks(comm, all_ces, args, verbose=True):
    """List operational day limits in the list of CES:s."""
    breaks = []
    if not args.do_daymaps:
        return breaks
    do_break = False
    nces = len(all_ces)
    for i in range(nces - 1):
        # If current and next CES are on different days, insert a break
        tz = args.timezone / 24
        start1 = all_ces[i][3]  # MJD start
        start2 = all_ces[i + 1][3]  # MJD start
        scan1 = all_ces[i][4]
        scan2 = all_ces[i + 1][4]
        if scan1 != scan2 and do_break:
            breaks.append(nces + i + 1)
            do_break = False
            continue
        day1 = int(start1 + tz)
        day2 = int(start2 + tz)
        if day1 != day2:
            if scan1 == scan2:
                # We want an entire CES, even if it crosses the day bound.
                # Wait until the scan number changes.
                do_break = True
            else:
                breaks.append(nces + i + 1)

    nbreak = len(breaks)
    if nbreak < comm.ngroups - 1:
        if comm.comm_world is None or comm.world_rank == 0:
            print(
                "WARNING: there are more process groups than observing days. "
                "Will try distributing by observation.",
                flush=True,
            )
        breaks = []
        for i in range(nces - 1):
            scan1 = all_ces[i][4]
            scan2 = all_ces[i + 1][4]
            if scan1 != scan2:
                breaks.append(nces + i + 1)
        nbreak = len(breaks)

    if nbreak != comm.ngroups - 1:
        raise RuntimeError(
            "Number of observing days ({}) does not match number of process "
            "groups ({}).".format(nbreak + 1, comm.ngroups)
        )
    return breaks


def _parse_line(line):
    """Parse one line of the schedule file"""
    if line.startswith("#"):
        return None

    fields = line.split()
    nfield = len(fields)
    if nfield == 22:
        # Deprecated prior to 2020-02 schedule format without boresight rotation field
        (
            start_date,
            start_time,
            stop_date,
            stop_time,
            mjdstart,
            mjdstop,
            name,
            azmin,
            azmax,
            el,
            rs,
            sun_el1,
            sun_az1,
            sun_el2,
            sun_az2,
            moon_el1,
            moon_az1,
            moon_el2,
            moon_az2,
            moon_phase,
            scan,
            subscan,
        ) = line.split()
        boresight_angle = 0
    else:
        # 2020-02 schedule format with boresight rotation field
        (
            start_date,
            start_time,
            stop_date,
            stop_time,
            mjdstart,
            mjdstop,
            boresight_angle,
            name,
            azmin,
            azmax,
            el,
            rs,
            sun_el1,
            sun_az1,
            sun_el2,
            sun_az2,
            moon_el1,
            moon_az1,
            moon_el2,
            moon_az2,
            moon_phase,
            scan,
            subscan,
        ) = line.split()
    start_time = start_date + " " + start_time
    stop_time = stop_date + " " + stop_time
    # Define season as a calendar year.  This can be
    # changed later and could even be in the schedule file.
    season = int(start_date.split("-")[0])
    try:
        start_time = dateutil.parser.parse(start_time + " +0000")
        stop_time = dateutil.parser.parse(stop_time + " +0000")
    except Exception:
        start_time = dateutil.parser.parse(start_time)
        stop_time = dateutil.parser.parse(stop_time)
    start_timestamp = start_time.timestamp()
    stop_timestamp = stop_time.timestamp()
    # useful metadata
    mindist_sun = min_sso_dist(
        *np.array([el, azmin, azmax, sun_el1, sun_az1, sun_el2, sun_az2]).astype(
            np.float
        )
    )
    mindist_moon = min_sso_dist(
        *np.array([el, azmin, azmax, moon_el1, moon_az1, moon_el2, moon_az2]).astype(
            np.float
        )
    )
    el_sun = max(float(sun_el1), float(sun_el2))
    return (
        start_timestamp,
        start_date,
        stop_timestamp,
        season,
        float(mjdstart),
        float(mjdstop),
        float(boresight_angle),
        name,
        float(azmin),
        float(azmax),
        float(el),
        int(scan),
        int(subscan),
        mindist_sun,
        mindist_moon,
        el_sun,
        rs.upper() == "R",
    )


@function_timer
def min_sso_dist(el, azmin, azmax, sso_el1, sso_az1, sso_el2, sso_az2):
    """Return a rough minimum angular distance between the bore sight
    and a solar system object"""
    sso_vec1 = hp.dir2vec(sso_az1, sso_el1, lonlat=True)
    sso_vec2 = hp.dir2vec(sso_az2, sso_el2, lonlat=True)
    az1 = azmin
    az2 = azmax
    if az2 < az1:
        az2 += 360
    n = 100
    az = np.linspace(az1, az2, n)
    el = np.ones(n) * el
    vec = hp.dir2vec(az, el, lonlat=True)
    dist1 = np.degrees(np.arccos(np.dot(sso_vec1, vec)))
    dist2 = np.degrees(np.arccos(np.dot(sso_vec2, vec)))
    return min(np.amin(dist1), np.amin(dist2))


@function_timer
def load_schedule(args, comm):
    """Load the observing schedule(s).

    Returns:
        schedules (list): List of tuples of the form
            (`site`, `all_ces`) where `all_ces` is
            a list of individual CES objects for `site`.
    """
    schedules = []
    timer0 = Timer()
    timer0.start()

    if comm.comm_world is None or comm.world_rank == 0:
        timer1 = Timer()
        isplit, nsplit = None, None
        if args.split_schedule is not None:
            isplit, nsplit = args.split_schedule.split(",")
            isplit = np.int(isplit)
            nsplit = np.int(nsplit)
            scan_counters = {}
        for fn in args.schedule.split(","):
            if not os.path.isfile(fn):
                raise RuntimeError("No such schedule file: {}".format(fn))
            timer1.start()
            with open(fn, "r") as f:
                while True:
                    line = f.readline()
                    if line.startswith("#"):
                        continue
                    (
                        site_name,
                        telescope_name,
                        site_lat,
                        site_lon,
                        site_alt,
                    ) = line.split()
                    site = Site(site_name, site_lat, site_lon, float(site_alt))
                    telescope = Telescope(telescope_name, site=site)
                    break
                all_ces = []
                last_name = None
                for line in f:
                    if line.startswith("#"):
                        continue
                    (
                        start_timestamp,
                        start_date,
                        stop_timestamp,
                        season,
                        mjdstart,
                        mjdstop,
                        boresight_angle,
                        name,
                        azmin,
                        azmax,
                        el,
                        scan,
                        subscan,
                        mindist_sun,
                        mindist_moon,
                        el_sun,
                        rising,
                    ) = _parse_line(line)
                    if nsplit:
                        # Only accept 1 / `nsplit` of the rising and setting
                        # scans in patch `name`.  Selection is performed
                        # during the first subscan.
                        if name != last_name:
                            if name not in scan_counters:
                                scan_counters[name] = {}
                            counter = scan_counters[name]
                            # Separate counters for rising and setting scans
                            if rising not in counter:
                                counter[rising] = 0
                            else:
                                counter[rising] += 1
                            iscan = counter[rising]
                        last_name = name
                        if iscan % nsplit != isplit:
                            continue
                    all_ces.append(
                        CES(
                            start_time=start_timestamp,
                            stop_time=stop_timestamp,
                            boresight_angle=boresight_angle,
                            name=name,
                            mjdstart=mjdstart,
                            scan=scan,
                            subscan=subscan,
                            azmin=azmin,
                            azmax=azmax,
                            el=el,
                            season=season,
                            start_date=start_date,
                            rising=rising,
                            mindist_sun=mindist_sun,
                            mindist_moon=mindist_moon,
                            el_sun=el_sun,
                        )
                    )
            schedules.append(Schedule(telescope, all_ces, sort=args.sort_schedule))
            timer1.report_clear("Load {} (sub)scans in {}".format(len(all_ces), fn))

    if comm.comm_world is not None:
        schedules = comm.comm_world.bcast(schedules)

    if comm.world_rank == 0:
        timer0.report_clear("Loading schedule(s)")
    return schedules


@function_timer
def load_weather(args, comm, schedules, verbose=False):
    """Load TOAST weather file(s) and attach them to the sites in the
     schedules.

    Args:
        schedules (iterable) :  list of observing schedules.
    """
    if args.weather is None:
        return
    timer = Timer()
    timer.start()

    if comm.world_rank == 0:
        weathers = []
        weatherdict = {}
        if args.weather == "SIM":
            weatherdict["SIM"] = Weather(None)
            weathers.append(weatherdict["SIM"])
        else:
            ftimer = Timer()
            for fname in args.weather.split(","):
                if fname not in weatherdict:
                    if not os.path.isfile(fname):
                        raise RuntimeError("No such weather file: {}".format(fname))
                    ftimer.start()
                    weatherdict[fname] = Weather(fname)
                    ftimer.stop()
                    ftimer.report_clear("Load {}".format(fname))
                weathers.append(weatherdict[fname])
    else:
        weathers = None

    if comm.comm_world is not None:
        weathers = comm.comm_world.bcast(weathers)
    if len(weathers) == 1 and len(schedules) > 1:
        weathers *= len(schedules)
    if len(weathers) != len(schedules):
        raise RuntimeError("Number of weathers must equal number of schedules or be 1.")

    for schedule, weather in zip(schedules, weathers):
        schedule.telescope.site.weather = weather

    if comm.world_rank == 0 and verbose:
        timer.report_clear("Loading weather")



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.


class Operator(object):
    """Base class for an operator that acts on collections of observations.

    An operator takes as input a toast.dist.Data object and modifies it in place.

    Args:
        None

    """

    def __init__(self):
        pass

    def exec(self, data):
        """Perform operations on a Data object.

        Args:
            data (toast.Data):  The distributed data.

        Returns:
            None

        """
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import re

from time import time

import numpy as np

from .._libtoast import filter_polynomial

from ..op import Operator

from ..timing import function_timer

from .. import qarray as qa


XAXIS, YAXIS, ZAXIS = np.eye(3)


class OpPolyFilter2D(Operator):
    """Operator to regress out 2D polynomials across the focal plane."""

    def __init__(
        self,
        order=1,
        pattern=r".*",
        name=None,
        common_flag_name=None,
        common_flag_mask=255,
        flag_name=None,
        flag_mask=255,
        poly_flag_mask=1,
        intervals="intervals",
    ):
        self._order = order
        self._nmode = (order + 1) * (order + 2) // 2
        self._pattern = pattern
        self._name = name
        self._common_flag_name = common_flag_name
        self._common_flag_mask = common_flag_mask
        self._flag_name = flag_name
        self._flag_mask = flag_mask
        self._poly_flag_mask = poly_flag_mask
        self._intervals = intervals

        # Call the parent class constructor.
        super().__init__()

    @function_timer
    def exec(self, data):
        """Apply the 2D polynomial filter to the signal.

        Args:
            data (toast.Data): The distributed data.

        """
        norder = self._order + 1
        nmode = self._nmode

        for obs in data.obs:
            t0 = time()
            t_template = 0
            t_get_norm = 0
            t_apply_norm = 0
            t_solve = 0
            t_clean = 0

            tod = obs["tod"]
            times = tod.local_times()
            comm = tod.grid_comm_row
            detectors = tod.detectors
            ndet = len(detectors)
            detector_index = {}
            pat = re.compile(self._pattern)

            ndet = 0
            for det in detectors:
                if pat.match(det) is None:
                    continue
                detector_index[det] = ndet
                ndet += 1
            # Number of detectors may limit the number of modes we can constrain
            nmode = min(self._nmode, ndet)

            focalplane = obs["focalplane"]

            detector_templates = np.zeros([ndet, nmode])
            mode = 0
            xorders = np.zeros(nmode)
            yorders = np.zeros(nmode)
            for order in range(norder):
                for yorder in range(order + 1):
                    xorder = order - yorder
                    xorders[mode] = xorder
                    yorders[mode] = yorder
                    mode += 1
                    if mode == nmode:
                        break
                if mode == nmode:
                    break

            for det in tod.local_dets:
                if det not in detector_index:
                    continue
                idet = detector_index[det]
                det_quat = focalplane[det]["quat"]
                x, y, z = qa.rotate(det_quat, ZAXIS)
                theta, phi = np.arcsin([x, y])
                detector_templates[idet] = theta ** xorders * phi ** yorders

            if self._intervals in obs:
                intervals = obs[self._intervals]
            else:
                intervals = None
            local_intervals = tod.local_intervals(intervals)
            if len(local_intervals) == 0:
                # No intervals to filter
                continue
            common_ref = tod.local_common_flags(self._common_flag_name)

            # Iterate over each interval

            for ival in local_intervals:
                ind = slice(ival.first, ival.last + 1)
                nsample = ival.last - ival.first + 1
                templates = np.zeros([ndet, nmode, nsample])
                proj = np.zeros([nmode, nsample])

                t1 = time()

                norms = np.zeros(nmode)

                for det in tod.local_dets:
                    if det not in detector_index:
                        continue
                    idet = detector_index[det]

                    ref = tod.local_signal(det, self._name)[ind]
                    flag_ref = tod.local_flags(det, self._flag_name)[ind]

                    flg = common_ref[ind] & self._common_flag_mask
                    flg |= flag_ref & self._flag_mask
                    mask = flg == 0

                    # We might want to remove the interval mean if the
                    # data were not already 1D-filtered
                    # ref -= np.mean(ref[mask])

                    template = detector_templates[idet]
                    templates[idet] = np.outer(template, mask)
                    proj += np.outer(template, ref * mask)
                    norms += template ** 2

                    del ref
                    del flag_ref

                t_template += time() - t1

                t1 = time()
                comm.allreduce(templates)
                comm.allreduce(proj)
                comm.allreduce(norms)
                good = norms != 0
                norms[good] = norms[good] ** -0.5
                t_get_norm += time() - t1

                t1 = time()
                templates = np.transpose(
                    templates, [1, 0, 2]
                ).copy()  # nmode x ndet x nsample
                for mode, norm in enumerate(norms):
                    if norm:
                        templates[mode] *= norm
                        proj[mode] *= norm
                t_apply_norm += time() - t1

                t1 = time()
                templates = np.transpose(
                    templates, [2, 1, 0]
                ).copy()  # nsample x ndet x nmode
                proj = proj.T.copy()  # nsample x nmode
                coeff = np.zeros([nsample, nmode])
                for isample in range(nsample):
                    if isample % comm.size != comm.rank:
                        continue
                    templatesT = templates[isample].T.copy()  # ndet x nmode
                    ccinv = np.dot(templatesT, templates[isample])
                    try:
                        cc = np.linalg.inv(ccinv)
                        coeff[isample] = np.dot(cc, proj[isample])
                    except np.linalg.LinAlgError:
                        coeff[isample] = 0
                comm.allreduce(coeff)
                t_solve += time() - t1

                t1 = time()

                """
                for isample in range(nsample):
                    if np.all(coeff[isample] == 0):
                        common_ref[isample + ival.first] |= self._poly_flag_mask
                        continue
                    for det in tod.local_dets:
                        if det not in detector_index:
                            continue
                        idet = detector_index[det]
                        ref = tod.local_signal(det, self._name)[ind]
                        ref[isample] -= np.dot(coeff[isample], templates[isample, idet])
                """

                for isample in range(nsample):
                    if np.all(coeff[isample] == 0):
                        common_ref[isample + ival.first] |= self._poly_flag_mask

                templates = np.transpose(
                    templates, [1, 2, 0]
                ).copy()  # ndet x nmode x nsample
                coeff = coeff.T.copy()  # nmode x nsample

                for det in tod.local_dets:
                    if det not in detector_index:
                        continue
                    idet = detector_index[det]
                    ref = tod.local_signal(det, self._name)[ind]
                    for mode in range(nmode):
                        ref -= coeff[mode] * templates[idet, mode]

                t_clean += time() - t1

            del common_ref

            """
            print(
                "Time per observation: {:.1f} s\n"
                "   templates : {:6.1f} s\n"
                "    get_norm : {:6.1f} s\n"
                "  apply_norm : {:6.1f} s\n"
                "       solve : {:6.1f} s\n"
                "       clean : {:6.1f} s".format(
                    time() - t0, t_template, t_get_norm, t_apply_norm, t_solve, t_clean
                ),
                flush=True,
            )
            """

        return


class OpPolyFilter(Operator):
    """Operator which applies polynomial filtering to the TOD.

    This applies polynomial filtering to the valid intervals of each TOD.

    Args:
        order (int):  Order of the filtering polynomial.
        pattern (str):  Regex pattern to match against detector names.
            Only detectors that match the pattern are filtered.
        name (str):  Name of the output signal cache object will be
            <name_in>_<detector>.  If the object exists, it is used as
            input.  Otherwise signal is read using the tod read method.
        common_flag_name (str):  Cache name of the output common flags.
            If it already exists, it is used.  Otherwise flags
            are read from the tod object and stored in the cache under
            common_flag_name.
        common_flag_mask (byte):  Bitmask to use when flagging data
           based on the common flags.
        flag_name (str):  Cache name of the output detector flags will
            be <flag_name>_<detector>.  If the object exists, it is
            used.  Otherwise flags are read from the tod object.
        flag_mask (byte):  Bitmask to use when flagging data
           based on the detector flags.
        poly_flag_mask (byte):  Bitmask to use when adding flags based
           on polynomial filter failures.
        intervals (str):  Name of the valid intervals in observation.

    """

    def __init__(
        self,
        order=1,
        pattern=r".*",
        name=None,
        common_flag_name=None,
        common_flag_mask=255,
        flag_name=None,
        flag_mask=255,
        poly_flag_mask=1,
        intervals="intervals",
    ):
        self._order = order
        self._pattern = pattern
        self._name = name
        self._common_flag_name = common_flag_name
        self._common_flag_mask = common_flag_mask
        self._flag_name = flag_name
        self._flag_mask = flag_mask
        self._poly_flag_mask = poly_flag_mask
        self._intervals = intervals

        # Call the parent class constructor.
        super().__init__()

    @function_timer
    def exec(self, data):
        """Apply the polynomial filter to the signal.

        Args:
            data (toast.Data): The distributed data.

        """
        for obs in data.obs:
            tod = obs["tod"]
            if self._intervals in obs:
                intervals = obs[self._intervals]
            else:
                intervals = None
            local_intervals = tod.local_intervals(intervals)
            if len(local_intervals) == 0:
                # No intervals to filter
                continue
            common_ref = tod.local_common_flags(self._common_flag_name)

            pat = re.compile(self._pattern)

            for det in tod.local_dets:
                # Test the detector pattern
                if pat.match(det) is None:
                    continue

                ref = tod.local_signal(det, self._name)
                flag_ref = tod.local_flags(det, self._flag_name)

                # Iterate over each interval

                local_starts = []
                local_stops = []
                for ival in local_intervals:
                    local_starts.append(ival.first)
                    local_stops.append(ival.last)

                local_starts = np.array(local_starts)
                local_stops = np.array(local_stops)

                flg = common_ref & self._common_flag_mask
                flg |= flag_ref & self._flag_mask

                filter_polynomial(self._order, flg, [ref], local_starts, local_stops)

                flag_ref[flg != 0] |= self._poly_flag_mask

                del ref
                del flag_ref

            del common_ref

        return



#!/usr/bin/env python3

# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

"""
This script runs a ground simulation and makes a map.
"""


import os

if "TOAST_STARTUP_DELAY" in os.environ:
    import numpy as np
    import time

    delay = np.float(os.environ["TOAST_STARTUP_DELAY"])
    wait = np.random.rand() * delay
    # print('Sleeping for {} seconds before importing TOAST'.format(wait),
    #      flush=True)
    time.sleep(wait)

import copy
import sys
import argparse
import traceback
import pickle

import numpy as np

from toast.mpi import get_world, Comm

from toast.dist import distribute_uniform, Data

from toast.utils import Logger, Environment, memreport

from toast.timing import function_timer, GlobalTimers, Timer, gather_timers
from toast.timing import dump as dump_timing

from toast.todmap import TODGround

from toast import pipeline_tools

# import warnings
# warnings.filterwarnings('error')
# warnings.simplefilter('ignore', ImportWarning)
# warnings.simplefilter('ignore', ResourceWarning)
# warnings.simplefilter('ignore', DeprecationWarning)
# warnings.filterwarnings("ignore", message="numpy.dtype size changed")
# warnings.filterwarnings("ignore", message="numpy.ufunc size changed")


def parse_arguments(comm):
    timer = Timer()
    timer.start()
    log = Logger.get()

    parser = argparse.ArgumentParser(
        description="Simulate ground-based boresight pointing.  Simulate "
        "atmosphere and make maps for some number of noise Monte Carlos.",
        fromfile_prefix_chars="@",
    )

    pipeline_tools.add_dist_args(parser)
    pipeline_tools.add_debug_args(parser)
    pipeline_tools.add_todground_args(parser)
    pipeline_tools.add_pointing_args(parser)
    pipeline_tools.add_polyfilter_args(parser)
    pipeline_tools.add_polyfilter2D_args(parser)
    pipeline_tools.add_groundfilter_args(parser)
    pipeline_tools.add_atmosphere_args(parser)
    pipeline_tools.add_noise_args(parser)
    pipeline_tools.add_gainscrambler_args(parser)
    pipeline_tools.add_madam_args(parser)
    pipeline_tools.add_mapmaker_args(parser)
    pipeline_tools.add_sky_map_args(parser)
    pipeline_tools.add_pysm_args(parser)
    pipeline_tools.add_sss_args(parser)
    pipeline_tools.add_tidas_args(parser)
    pipeline_tools.add_spt3g_args(parser)
    pipeline_tools.add_mc_args(parser)

    parser.add_argument(
        "--outdir", required=False, default="out", help="Output directory"
    )

    parser.add_argument(
        "--madam",
        required=False,
        action="store_true",
        help="Use libmadam for map-making",
        dest="use_madam",
    )
    parser.add_argument(
        "--no-madam",
        required=False,
        action="store_false",
        help="Do not use libmadam for map-making [default]",
        dest="use_madam",
    )
    parser.set_defaults(use_madam=False)

    parser.add_argument(
        "--focalplane",
        required=False,
        default=None,
        help="Pickle file containing a dictionary of detector "
        "properties.  The keys of this dict are the detector "
        "names, and each value is also a dictionary with keys "
        '"quat" (4 element ndarray), "fwhm" (float, arcmin), '
        '"fknee" (float, Hz), "alpha" (float), and '
        '"NET" (float).',
    )
    parser.add_argument(
        "--freq",
        required=True,
        help="Comma-separated list of frequencies with identical focal planes."
        "  They override the bandpasses in the focalplane for the purpose of"
        " scaling the atmospheric signal but not for simulating the sky signal.",
    )

    try:
        args = parser.parse_args()
    except SystemExit:
        sys.exit(0)

    if args.tidas is not None:
        if not tidas_available:
            raise RuntimeError("TIDAS not found- cannot export")

    if args.spt3g is not None:
        if not spt3g_available:
            raise RuntimeError("SPT3G not found- cannot export")

    if len(args.freq.split(",")) != 1:
        # Multi frequency run.  We don't support multiple copies of
        # scanned signal.
        if args.input_map:
            raise RuntimeError(
                "Multiple frequencies are not supported when scanning from a map"
            )

    if args.simulate_atmosphere and args.weather is None:
        raise RuntimeError("Cannot simulate atmosphere without a TOAST weather file")

    if comm.world_rank == 0:
        log.info("All parameters:")
        for ag in vars(args):
            log.info("{} = {}".format(ag, getattr(args, ag)))

    if args.group_size:
        comm = Comm(groupsize=args.group_size)

    if comm.world_rank == 0:
        os.makedirs(args.outdir, exist_ok=True)

    timer.stop()
    if comm.world_rank == 0:
        timer.report("Parsed parameters")

    return args, comm


@function_timer
def load_focalplanes(args, comm, schedules):
    """Attach a focalplane to each of the schedules.

    Args:
        schedules (list) :  List of Schedule instances.
            Each schedule has two members, telescope
            and ceslist, a list of CES objects.
    Returns:
        detweights (dict) : Inverse variance noise weights for every
            detector across all focal planes. In [K_CMB^-2].
            They can be used to bin the TOD.
    """
    timer = Timer()
    timer.start()

    # Load focalplane information

    focalplanes = []
    if comm.world_rank == 0:
        for fpfile in args.focalplane.split(","):
            focalplanes.append(
                pipeline_tools.Focalplane(
                    fname_pickle=fpfile,
                    sample_rate=args.sample_rate,
                    radius_deg=args.focalplane_radius_deg,
                )
            )
    if comm.comm_world is not None:
        focalplanes = comm.comm_world.bcast(focalplanes)

    if len(focalplanes) == 1 and len(schedules) > 1:
        focalplanes *= len(schedules)
    if len(focalplanes) != len(schedules):
        raise RuntimeError(
            "Number of focalplanes must equal number of schedules or be 1."
        )

    # Append a focal plane and telescope to each entry in the schedules
    # list and assemble a detector weight dictionary that represents all
    # detectors in all focalplanes
    detweights = {}
    for schedule, focalplane in zip(schedules, focalplanes):
        schedule.telescope.focalplane = focalplane
        detweights.update(schedule.telescope.focalplane.detweights)

    timer.stop()
    if comm.world_rank == 0:
        timer.report("Loading focalplanes")
    return detweights


@function_timer
def create_observation(args, comm, telescope, ces, verbose=True):
    """Create a TOAST observation.

    Create an observation for the CES scan

    Args:
        args :  argparse arguments
        comm :  TOAST communicator
        ces (CES) :  One constant elevation scan

    """
    focalplane = telescope.focalplane
    site = telescope.site
    weather = site.weather
    noise = focalplane.noise
    totsamples = int((ces.stop_time - ces.start_time) * args.sample_rate)

    # create the TOD for this observation

    if comm.comm_group is not None:
        ndetrank = comm.comm_group.size
    else:
        ndetrank = 1

    if args.el_nod_deg and (ces.subscan == 0 or args.el_nod_every_scan):
        el_nod = args.el_nod_deg
    else:
        el_nod = None

    try:
        tod = TODGround(
            comm.comm_group,
            focalplane.detquats,
            totsamples,
            detranks=ndetrank,
            boresight_angle=ces.boresight_angle,
            firsttime=ces.start_time,
            rate=args.sample_rate,
            site_lon=site.lon,
            site_lat=site.lat,
            site_alt=site.alt,
            azmin=ces.azmin,
            azmax=ces.azmax,
            el=ces.el,
            el_nod=el_nod,
            start_with_elnod=args.start_with_el_nod,
            end_with_elnod=args.end_with_el_nod,
            el_mod_step=args.el_mod_step_deg,
            el_mod_rate=args.el_mod_rate_hz,
            el_mod_amplitude=args.el_mod_amplitude_deg,
            el_mod_sine=args.el_mod_sine,
            scanrate=args.scan_rate,
            scanrate_el=args.scan_rate_el,
            scan_accel=args.scan_accel,
            scan_accel_el=args.scan_accel_el,
            cosecant_modulation=args.scan_cosecant_modulate,
            CES_start=None,
            CES_stop=None,
            sun_angle_min=args.sun_angle_min,
            coord=args.coord,
            sampsizes=None,
            report_timing=args.debug,
            hwprpm=args.hwp_rpm,
            hwpstep=args.hwp_step_deg,
            hwpsteptime=args.hwp_step_time_s,
        )
    except RuntimeError as e:
        raise RuntimeError(
            'Failed to create TOD for {}-{}-{}: "{}"'
            "".format(ces.name, ces.scan, ces.subscan, e)
        )

    # Create the observation

    obs = {}
    obs["name"] = "CES-{}-{}-{}-{}-{}".format(
        site.name, telescope.name, ces.name, ces.scan, ces.subscan
    )
    obs["tod"] = tod
    obs["baselines"] = None
    obs["noise"] = copy.deepcopy(noise)
    obs["id"] = int(ces.mjdstart * 10000)
    obs["intervals"] = tod.subscans
    obs["site"] = site
    obs["site_name"] = site.name
    obs["site_id"] = site.id
    obs["altitude"] = site.alt
    obs["weather"] = site.weather
    obs["telescope"] = telescope
    obs["telescope_name"] = telescope.name
    obs["telescope_id"] = telescope.id
    obs["focalplane"] = telescope.focalplane.detector_data
    obs["fpradius"] = telescope.focalplane.radius
    obs["start_time"] = ces.start_time
    obs["season"] = ces.season
    obs["date"] = ces.start_date
    obs["MJD"] = ces.mjdstart
    obs["rising"] = ces.rising
    obs["mindist_sun"] = ces.mindist_sun
    obs["mindist_moon"] = ces.mindist_moon
    obs["el_sun"] = ces.el_sun
    return obs


@function_timer
def create_observations(args, comm, schedules):
    """Create and distribute TOAST observations for every CES in
    schedules.

    Args:
        schedules (iterable) :  a list of Schedule objects.
    """
    log = Logger.get()
    timer = Timer()
    timer.start()

    data = Data(comm)

    # Loop over the schedules, distributing each schedule evenly across
    # the process groups.  For now, we'll assume that each schedule has
    # the same number of operational days and the number of process groups
    # matches the number of operational days.  Relaxing these constraints
    # will cause the season break to occur on different process groups
    # for different schedules and prevent splitting the communicator.

    for schedule in schedules:

        telescope = schedule.telescope
        all_ces = schedule.ceslist
        nces = len(all_ces)

        breaks = pipeline_tools.get_breaks(comm, all_ces, args)

        groupdist = distribute_uniform(nces, comm.ngroups, breaks=breaks)
        group_firstobs = groupdist[comm.group][0]
        group_numobs = groupdist[comm.group][1]

        for ices in range(group_firstobs, group_firstobs + group_numobs):
            obs = create_observation(args, comm, telescope, all_ces[ices])
            data.obs.append(obs)

    if comm.comm_world is None or comm.comm_group.rank == 0:
        log.info("Group # {:4} has {} observations.".format(comm.group, len(data.obs)))

    if len(data.obs) == 0:
        raise RuntimeError(
            "Too many tasks. Every MPI task must "
            "be assigned to at least one observation."
        )

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    timer.stop()
    if comm.world_rank == 0:
        timer.report("Simulated scans")

    # Split the data object for each telescope for separate mapmaking.
    # We could also split by site.

    if len(schedules) > 1:
        telescope_data = data.split("telescope_name")
        if len(telescope_data) == 1:
            # Only one telescope available
            telescope_data = []
    else:
        telescope_data = []
    telescope_data.insert(0, ("all", data))
    return data, telescope_data


def setup_sigcopy(args):
    """Determine if an extra copy of the atmospheric signal is needed.

    When we simulate multichroic focal planes, the frequency-independent
    part of the atmospheric noise is simulated first and then the
    frequency scaling is applied to a copy of the atmospheric noise.
    """
    if len(args.freq.split(",")) == 1:
        totalname = "total"
        totalname_freq = "total"
    else:
        totalname = "total"
        totalname_freq = "total_freq"

    return totalname, totalname_freq


@function_timer
def setup_output(args, comm, mc, freq):
    outpath = "{}/{:08}/{:03}".format(args.outdir, mc, int(freq))
    if comm.world_rank == 0:
        os.makedirs(outpath, exist_ok=True)
    return outpath


def main():
    log = Logger.get()
    gt = GlobalTimers.get()
    gt.start("toast_ground_sim (total)")
    timer0 = Timer()
    timer0.start()

    mpiworld, procs, rank, comm = pipeline_tools.get_comm()

    args, comm = parse_arguments(comm)

    if args.use_madam:
        # Initialize madam parameters
        madampars = pipeline_tools.setup_madam(args)

    # Load and broadcast the schedule file

    schedules = pipeline_tools.load_schedule(args, comm)

    # Load the weather and append to schedules

    pipeline_tools.load_weather(args, comm, schedules)

    # load or simulate the focalplane

    detweights = load_focalplanes(args, comm, schedules)

    # Create the TOAST data object to match the schedule.  This will
    # include simulating the boresight pointing.

    data, telescope_data = create_observations(args, comm, schedules)

    # Split the communicator for day and season mapmaking

    time_comms = pipeline_tools.get_time_communicators(args, comm, data)

    # Expand boresight quaternions into detector pointing weights and
    # pixel numbers

    pipeline_tools.expand_pointing(args, comm, data)

    # Optionally rewrite the noise PSD:s in each observation to include
    # elevation-dependence

    pipeline_tools.get_elevation_noise(args, comm, data)

    # Purge the pointing if we are NOT going to export the
    # data to a TIDAS volume
    if (args.tidas is None) and (args.spt3g is None):
        for ob in data.obs:
            tod = ob["tod"]
            tod.free_radec_quats()

    # Prepare auxiliary information for distributed map objects

    if args.pysm_model:
        focalplanes = [s.telescope.focalplane.detector_data for s in schedules]
        signalname = pipeline_tools.simulate_sky_signal(
            args, comm, data, focalplanes, "signal"
        )
    else:
        signalname = pipeline_tools.scan_sky_signal(args, comm, data, "signal")

    # Set up objects to take copies of the TOD at appropriate times

    totalname, totalname_freq = setup_sigcopy(args)

    # Loop over Monte Carlos

    firstmc = args.MC_start
    nsimu = args.MC_count

    freqs = [float(freq) for freq in args.freq.split(",")]
    nfreq = len(freqs)

    for mc in range(firstmc, firstmc + nsimu):

        pipeline_tools.simulate_atmosphere(args, comm, data, mc, totalname)

        # Loop over frequencies with identical focal planes and identical
        # atmospheric noise.

        for ifreq, freq in enumerate(freqs):

            if comm.world_rank == 0:
                log.info(
                    "Processing frequency {}GHz {} / {}, MC = {}".format(
                        freq, ifreq + 1, nfreq, mc
                    )
                )

            # Make a copy of the atmosphere so we can scramble the gains and apply
            # frequency-dependent scaling.
            pipeline_tools.copy_signal(args, comm, data, totalname, totalname_freq)

            pipeline_tools.scale_atmosphere_by_frequency(
                args, comm, data, freq=freq, mc=mc, cache_name=totalname_freq
            )

            pipeline_tools.update_atmospheric_noise_weights(args, comm, data, freq, mc)

            # Add previously simulated sky signal to the atmospheric noise.

            pipeline_tools.add_signal(
                args, comm, data, totalname_freq, signalname, purge=(nsimu == 1)
            )

            mcoffset = ifreq * 1000000

            pipeline_tools.simulate_noise(
                args, comm, data, mc + mcoffset, totalname_freq
            )

            pipeline_tools.simulate_sss(args, comm, data, mc + mcoffset, totalname_freq)

            pipeline_tools.scramble_gains(
                args, comm, data, mc + mcoffset, totalname_freq
            )

            if (mc == firstmc) and (ifreq == 0):
                # For the first realization and frequency, optionally
                # export the timestream data.
                pipeline_tools.output_tidas(args, comm, data, totalname)
                pipeline_tools.output_spt3g(args, comm, data, totalname)

            outpath = setup_output(args, comm, mc + mcoffset, freq)

            # Bin and destripe maps

            if args.use_madam:
                pipeline_tools.apply_madam(
                    args,
                    comm,
                    data,
                    madampars,
                    outpath,
                    detweights,
                    totalname_freq,
                    freq=freq,
                    time_comms=time_comms,
                    telescope_data=telescope_data,
                    first_call=(mc == firstmc),
                )
            else:
                pipeline_tools.apply_mapmaker(
                    args,
                    comm,
                    data,
                    outpath,
                    totalname_freq,
                    time_comms=time_comms,
                    telescope_data=telescope_data,
                    first_call=(mc == firstmc),
                )

            if (
                args.apply_polyfilter
                or args.apply_polyfilter2D
                or args.apply_groundfilter
            ):

                # Filter signal

                pipeline_tools.apply_polyfilter(args, comm, data, totalname_freq)

                pipeline_tools.apply_polyfilter2D(args, comm, data, totalname_freq)

                pipeline_tools.apply_groundfilter(args, comm, data, totalname_freq)

                # Bin filtered maps

                if args.use_madam:
                    pipeline_tools.apply_madam(
                        args,
                        comm,
                        data,
                        madampars,
                        outpath,
                        detweights,
                        totalname_freq,
                        freq=freq,
                        time_comms=time_comms,
                        telescope_data=telescope_data,
                        first_call=False,
                        extra_prefix="filtered",
                        bin_only=True,
                    )
                else:
                    pipeline_tools.apply_mapmaker(
                        args,
                        comm,
                        data,
                        outpath,
                        totalname_freq,
                        time_comms=time_comms,
                        telescope_data=telescope_data,
                        first_call=False,
                        extra_prefix="filtered",
                        bin_only=True,
                    )

    gt.stop_all()
    if mpiworld is not None:
        mpiworld.barrier()
    timer = Timer()
    timer.start()
    alltimers = gather_timers(comm=mpiworld)
    if comm.world_rank == 0:
        out = os.path.join(args.outdir, "timing")
        dump_timing(alltimers, out)
        timer.stop()
        timer.report("Gather and dump timing info")
        timer0.report_clear("toast_ground_sim.py")
    return


if __name__ == "__main__":
    try:
        main()
    except Exception:
        # We have an unhandled exception on at least one process.  Print a stack
        # trace for this process and then abort so that all processes terminate.
        mpiworld, procs, rank = get_world()
        if procs == 1:
            raise
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        lines = ["Proc {}: {}".format(rank, x) for x in lines]
        print("".join(lines), flush=True)
        if mpiworld is not None:
            mpiworld.Abort(6)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import numpy as np

# from contextlib import contextmanager

from ..mpi import Comm

from ..dist import Data

from .. import qarray as qa


ZAXIS = np.array([0, 0, 1.0])


# These are helper routines for common operations used in the unit tests.


def create_outdir(mpicomm, subdir=None):
    """Create the top level output directory and per-test subdir.

    Args:
        mpicomm (MPI.Comm): the MPI communicator (or None).
        subdir (str): the sub directory for this test.

    Returns:
        str: full path to the test subdir if specified, else the top dir.

    """
    pwd = os.path.abspath(".")
    testdir = os.path.join(pwd, "toast_test_output")
    retdir = testdir
    if subdir is not None:
        retdir = os.path.join(testdir, subdir)
    if (mpicomm is None) or (mpicomm.rank == 0):
        if not os.path.isdir(testdir):
            os.mkdir(testdir)
        if not os.path.isdir(retdir):
            os.mkdir(retdir)
    if mpicomm is not None:
        mpicomm.barrier()
    return retdir


def create_comm(mpicomm):
    """Create a toast communicator.

    Use the specified MPI communicator to attempt to create 2 process groups.
    If less than 2 processes are used, create a single process group.

    Args:
        mpicomm (MPI.Comm): the MPI communicator (or None).

    Returns:
        toast.Comm: the 2-level toast communicator.

    """
    toastcomm = None
    if mpicomm is None:
        toastcomm = Comm(world=mpicomm)
    else:
        worldsize = mpicomm.size
        groupsize = 1
        if worldsize >= 2:
            groupsize = worldsize // 2
        toastcomm = Comm(world=mpicomm, groupsize=groupsize)
    return toastcomm


def create_distdata(mpicomm, obs_per_group=1):
    """Create a toast communicator and distributed data object.

    Use the specified MPI communicator to attempt to create 2 process groups,
    each with some observations.

    Args:
        mpicomm (MPI.Comm): the MPI communicator (or None).
        obs_per_group (int): the number of observations assigned to each group.

    Returns:
        toast.Data: the distributed data with named observations (but no TOD).

    """
    toastcomm = create_comm(mpicomm)
    data = Data(toastcomm)
    for obs in range(obs_per_group):
        ob = {}
        ob["name"] = "test-{}-{}".format(toastcomm.group, obs)
        ob["id"] = obs_per_group * toastcomm.group + obs
        data.obs.append(ob)
    return data


def uniform_chunks(samples, nchunk=100):
    """Divide some number of samples into chunks.

    This is often needed when constructing a TOD class, and usually we want
    the number of chunks to be larger than any number of processes we might
    be using for the unit tests.

    Args:
        samples (int): The number of samples.
        nchunk (int): The number of chunks to create.

    Returns:
        array: This list of chunk sizes.

    """
    chunksize = samples // nchunk
    chunks = np.ones(nchunk, dtype=np.int64)
    chunks *= chunksize
    remain = samples - (nchunk * chunksize)
    for r in range(remain):
        chunks[r] += 1
    return chunks


def boresight_focalplane(
    ndet, samplerate=1.0, epsilon=0.0, net=1.0, fmin=0.0, alpha=1.0, fknee=0.05
):
    """Create a set of detectors at the boresight.

    This creates multiple detectors at the boresight, oriented in evenly
    spaced increments from zero to PI.

    Args:
        ndet (int): the number of detectors.


    Returns:
        (tuple): names(list), quat(dict), fmin(dict), rate(dict), fknee(dict),
            alpha(dict), netd(dict)

    """
    names = ["d{:02d}".format(x) for x in range(ndet)]
    pol = {"d{:02d}".format(x): (x * np.pi / ndet) for x in range(ndet)}

    quat = {
        "d{:02d}".format(x): qa.rotation(ZAXIS, pol["d{:02d}".format(x)])
        for x in range(ndet)
    }

    det_eps = {"d{:02d}".format(x): epsilon for x in range(ndet)}

    det_fmin = {"d{:02d}".format(x): fmin for x in range(ndet)}
    det_rate = {"d{:02d}".format(x): samplerate for x in range(ndet)}
    det_alpha = {"d{:02d}".format(x): alpha for x in range(ndet)}
    det_net = {"d{:02d}".format(x): net for x in range(ndet)}

    det_fknee = None
    if np.isscalar(fknee):
        det_fknee = {"d{:02d}".format(x): fknee for x in range(ndet)}
    else:
        # This must be an array or list of correct length
        if len(fknee) != ndet:
            raise RuntimeError("length of knee frequencies must equal ndet")
        det_fknee = {"d{:02d}".format(x): y for x, y in zip(range(ndet), fknee)}

    return names, quat, det_eps, det_rate, det_net, det_fmin, det_fknee, det_alpha


def create_weather(outfile):
    from astropy.table import Table
    import astropy.io.fits as af

    nstep = 101
    TQI = [np.linspace(0, 0.5, nstep) for x in range(24)]
    TQL = [np.linspace(0, 0.1, nstep) for x in range(24)]
    TQV = [np.linspace(1, 12, nstep) for x in range(24)]
    QV10M = [np.linspace(0.001, 0.007, nstep) for x in range(24)]
    PS = [np.linspace(58600, 59000, nstep) for x in range(24)]
    TS = [np.linspace(270, 280, nstep) for x in range(24)]
    T10M = [np.linspace(270, 280, nstep) for x in range(24)]
    U10M = [np.linspace(-2, 8, nstep) for x in range(24)]
    V10M = [np.linspace(-1, 3, nstep) for x in range(24)]

    hdus = af.HDUList([af.PrimaryHDU()])

    for mon in range(12):
        tab = Table(
            [TQI, TQL, TQV, QV10M, PS, TS, T10M, U10M, V10M],
            names=("TQI", "TQL", "TQV", "QV10M", "PS", "TS", "T10M", "U10M", "V10M"),
            meta={
                "PROBSTRT": 0.0,
                "PROBSTOP": 1.0,
                "PROBSTEP": 0.01,
                "NSTEP": nstep,
                "MONTH": mon,
            },
        )
        hdus.append(af.table_to_hdu(tab))

    hdus.writeto(outfile, overwrite=True)

    return


#
# @contextmanager
# def mpi_guard(comm=MPI.COMM_WORLD):
#     """Ensure that if one MPI process raises an exception, all of them do.
#
#     Args:
#         comm (mpi4py.MPI.Comm): The MPI communicator.
#
#     """
#     failed = 0
#     print(comm.rank, ": guard: enter", flush=True)
#     try:
#         print(comm.rank, ": guard: yield", flush=True)
#         yield
#     except:
#         print(comm.rank, ": guard: except", flush=True)
#         msg = "Exception on process {}:\n".format(comm.rank)
#         exc_type, exc_value, exc_traceback = sys.exc_info()
#         lines = traceback.format_exception(exc_type, exc_value,
#             exc_traceback)
#         msg += "\n".join(lines)
#         print(msg, flush=True)
#         failed = 1
#         print(comm.rank, ": guard: except done", flush=True)
#
#     print(comm.rank, ": guard: failcount reduce", flush=True)
#     failcount = comm.allreduce(failed, op=MPI.SUM)
#     if failcount > 0:
#         raise RuntimeError("One or more MPI processes raised an exception")
#     print(comm.rank, ": guard: done", flush=True)
#
#     return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from ..mpi import MPI

import numpy as np

from numpy.polynomial.chebyshev import chebval

from .._libtoast import bin_templates, add_templates, chebyshev

from ..op import Operator

from ..utils import Logger

from ..timing import function_timer


class OpGroundFilter(Operator):
    """Operator which applies ground template filtering to constant
    elevation scans.

    Args:
        name (str):  Name of the output signal cache object will be
            <name_in>_<detector>.  If the object exists, it is used as
            input.  Otherwise signal is read using the tod read method.
        common_flag_name (str):  Cache name of the output common flags.
            If it already exists, it is used.  Otherwise flags
            are read from the tod object and stored in the cache under
            common_flag_name.
        common_flag_mask (byte):  Bitmask to use when flagging data
           based on the common flags.
        flag_name (str):  Cache name of the output detector flags will
            be <flag_name>_<detector>.  If the object exists, it is
            used.  Otherwise flags are read from the tod object.
        flag_mask (byte):  Bitmask to use when flagging data
           based on the detector flags.
        ground_flag_mask (byte):  Bitmask to use when adding flags based
           on ground filter failures.
        trend_order (int):  Order of a Chebyshev polynomial to fit along
            with the ground template
        filter_order (int):  Order of a Chebyshev polynomial to fit as a
            function of azimuth
        detrend (bool):  Subtract the fitted trend along with the
             ground template
        intervals (str):  Name of the valid intervals in observation
        split_template (bool):  Apply a different template for left and
             right scans

    """

    def __init__(
        self,
        name=None,
        common_flag_name=None,
        common_flag_mask=255,
        flag_name=None,
        flag_mask=255,
        ground_flag_mask=1,
        trend_order=5,
        filter_order=5,
        detrend=False,
        intervals="intervals",
        split_template=False,
    ):
        self._name = name
        self._common_flag_name = common_flag_name
        self._common_flag_mask = common_flag_mask
        self._flag_name = flag_name
        self._flag_mask = flag_mask
        self._ground_flag_mask = ground_flag_mask
        self._trend_order = trend_order
        self._filter_order = filter_order
        self._detrend = detrend
        self._intervals = intervals
        self._split_template = split_template

        # Call the parent class constructor.
        super().__init__()

    @function_timer
    def build_templates(self, tod, obs):
        """Construct the local ground template hierarchy"""
        if self._intervals in obs:
            intervals = obs[self._intervals]
        else:
            intervals = None
        local_intervals = tod.local_intervals(intervals)

        # Construct trend templates.  Full domain for x is [-1, 1]

        my_offset, my_nsamp = tod.local_samples
        nsamp_tot = tod.total_samples
        x = np.arange(my_offset, my_offset + my_nsamp) / nsamp_tot * 2 - 1

        # Do not include the offset in the trend.  It will be part of
        # of the ground template
        # cheby_trend = chebval(x, np.eye(self._trend_order + 1), tensor=True)[1:]
        cheby_trend = np.zeros([self._trend_order, x.size])
        chebyshev(x, cheby_trend, 1, self._trend_order + 1)

        try:
            (azmin, azmax, _, _) = tod.scan_range
            az = tod.read_boresight_az()
        except Exception as e:
            raise RuntimeError(
                "Failed to get boresight azimuth from TOD.  Perhaps it is "
                'not ground TOD? "{}"'.format(e)
            )

        # The azimuth vector is assumed to be arranged so that the
        # azimuth increases monotonously even across the zero meridian.

        phase = (az - azmin) / (azmax - azmin) * 2 - 1
        nfilter = self._filter_order + 1
        # cheby_templates = chebval(phase, np.eye(nfilter), tensor=True)
        cheby_templates = np.zeros([nfilter, phase.size])
        chebyshev(phase, cheby_templates, 0, nfilter)
        if not self._split_template:
            cheby_filter = cheby_templates
        else:
            # Create separate templates for alternating scans
            common_ref = tod.local_common_flags(self._common_flag_name)
            cheby_filter = []
            mask1 = common_ref & tod.LEFTRIGHT_SCAN == 0
            mask2 = common_ref & tod.RIGHTLEFT_SCAN == 0
            for template in cheby_templates:
                for mask in mask1, mask2:
                    temp = template.copy()
                    temp[mask] = 0
                    cheby_filter.append(temp)
            del common_ref
            cheby_filter = np.vstack(cheby_filter)

        templates = np.vstack([cheby_trend, cheby_filter])

        return templates, cheby_trend, cheby_filter

    @function_timer
    def bin_templates(self, ref, templates, good, invcov, proj):

        # Bin the local data

        for i in range(ntemplate):
            temp = templates[i] * good
            proj[i] = np.dot(temp, ref)
            for j in range(i, ntemplate):
                invcov[i, j] = np.dot(temp, templates[j])
                # Symmetrize invcov
                if i != j:
                    invcov[j, i] = invcov[i, j]
            del temp

        return

    @function_timer
    def fit_templates(self, tod, det, templates, ref, good):
        log = Logger.get()
        comm = tod.mpicomm
        rank = 0
        if comm is not None:
            rank = comm.rank
        detranks, sampranks = tod.grid_size

        if good is not None:
            ngood = np.sum(good)
        else:
            ngood = 0
        ngood_tot = None
        if comm is None:
            ngood_tot = ngood
        else:
            ngood_tot = comm.allreduce(ngood)
        if ngood_tot == 0:
            return None

        ntemplate = len(templates)
        invcov = np.zeros([ntemplate, ntemplate])
        proj = np.zeros(ntemplate)
        if ref is not None:
            # self.bin_templates(ref, templates, good, invcov, proj)
            bin_templates(ref, templates, good.astype(np.uint8), invcov, proj)

        if sampranks > 1:
            # Reduce the binned data.  The detector signals is
            # distributed across the group communicator.
            comm.Allreduce(MPI.IN_PLACE, invcov, op=MPI.SUM)
            comm.Allreduce(MPI.IN_PLACE, proj, op=MPI.SUM)

        # Assemble the joint template
        if ref is not None:
            try:
                cov = np.linalg.inv(invcov)
                coeff = np.dot(cov, proj)
            except np.linalg.LinAlgError as e:
                msg = 'linalg.inv failed with "{}"'.format(e)
                log.warning(msg)
                # np.linalg.lstsq will find a least squares minimum
                # even if the covariance matrix is not invertible
                coeff = np.linalg.lstsq(invcov, proj, rcond=1e-30)[0]
                if np.any(np.isnan(coeff)) or np.std(coeff) < 1e-30:
                    raise RuntimeError("lstsq FAILED")
        else:
            coeff = None

        return coeff

    @function_timer
    def subtract_templates(self, ref, good, coeff, cheby_trend, cheby_filter):
        # Trend
        if self._detrend:
            trend = np.zeros_like(ref)
            # for cc, template in zip(coeff[: self._trend_order], cheby_trend):
            #    trend += cc * template
            add_templates(trend, cheby_trend, coeff[: self._trend_order])
            ref[good] -= trend[good]
        # Ground template
        grtemplate = np.zeros_like(ref)
        # for cc, template in zip(coeff[self._trend_order :], cheby_filter):
        #    grtemplate += cc * template
        add_templates(grtemplate, cheby_filter, coeff[self._trend_order :])
        ref[good] -= grtemplate[good]
        ref[np.logical_not(good)] = 0
        return

    @function_timer
    def exec(self, data):
        """Apply the ground filter to the signal.

        Args:
            data (toast.Data): The distributed data.

        """
        # Each group loops over its own CES:es
        for obs in data.obs:
            tod = obs["tod"]

            # Cache the output common flags
            common_ref = tod.local_common_flags(self._common_flag_name)

            templates, cheby_trend, cheby_filter = self.build_templates(tod, obs)

            for det in tod.detectors:
                if det in tod.local_dets:
                    ref = tod.local_signal(det, self._name)
                    flag_ref = tod.local_flags(det, self._flag_name)
                    good = np.logical_and(
                        common_ref & self._common_flag_mask == 0,
                        flag_ref & self._flag_mask == 0,
                    )
                    del flag_ref
                else:
                    ref = None
                    good = None

                coeff = self.fit_templates(tod, det, templates, ref, good)
                if coeff is not None:
                    self.subtract_templates(ref, good, coeff, cheby_trend, cheby_filter)

                del ref

            del common_ref

        return
